"""
Manage nspawn containers

.. versionadded:: 2015.8.0

`systemd-nspawn(1)`__ is a tool used to manage lightweight namespace
containers. This execution module provides several functions to help manage
these containers.

.. __: http://www.freedesktop.org/software/systemd/man/systemd-nspawn.html

Minions running systemd >= 219 will place new containers in
``/var/lib/machines``, while those running systemd < 219 will place them in
``/var/lib/container``.

.. note:

    ``nsenter(1)`` is required to run commands within containers. It should
    already be present on any systemd host, as part of the **util-linux**
    package.
"""

import errno
import functools
import logging
import os
import re
import shutil
import tempfile
import time

import salt.defaults.exitcodes
import salt.utils.args
import salt.utils.functools
import salt.utils.path
import salt.utils.systemd
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

__func_alias__ = {
    "list_": "list",
}
__virtualname__ = "nspawn"
SEED_MARKER = "/nspawn.initial_seed"
WANT = "/etc/systemd/system/multi-user.target.wants/systemd-nspawn@{0}.service"
EXEC_DRIVER = "nsenter"


def __virtual__():
    """
    Only work on systems that have been booted with systemd
    """
    if __grains__["kernel"] == "Linux" and salt.utils.systemd.booted(__context__):
        if salt.utils.systemd.version() is None:
            log.error("nspawn: Unable to determine systemd version")
        else:
            return __virtualname__
    return (
        False,
        "The nspawn execution module failed to load: "
        "only work on systems that have been booted with systemd.",
    )


def _sd_version():
    """
    Returns __context__.get('systemd.version', 0), avoiding duplication of the
    call to dict.get and making it easier to change how we handle this context
    var in the future
    """
    return salt.utils.systemd.version(__context__)


def _ensure_exists(wrapped):
    """
    Decorator to ensure that the named container exists.
    """

    @functools.wraps(wrapped)
    def check_exists(name, *args, **kwargs):
        if not exists(name):
            raise CommandExecutionError(f"Container '{name}' does not exist")
        return wrapped(name, *args, **salt.utils.args.clean_kwargs(**kwargs))

    return check_exists


def _root(name="", all_roots=False):
    """
    Return the container root directory. Starting with systemd 219, new
    images go into /var/lib/machines.
    """
    if _sd_version() >= 219:
        if all_roots:
            return [
                os.path.join(x, name)
                for x in ("/var/lib/machines", "/var/lib/container")
            ]
        else:
            return os.path.join("/var/lib/machines", name)
    else:
        ret = os.path.join("/var/lib/container", name)
        if all_roots:
            return [ret]
        else:
            return ret


def _make_container_root(name):
    """
    Make the container root directory
    """
    path = _root(name)
    if os.path.exists(path):
        __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
        raise CommandExecutionError(f"Container {name} already exists")
    else:
        try:
            os.makedirs(path)
            return path
        except OSError as exc:
            raise CommandExecutionError(
                f"Unable to make container root directory {name}: {exc}"
            )


def _build_failed(dst, name):
    try:
        __context__["retcode"] = salt.defaults.exitcodes.SALT_BUILD_FAIL
        shutil.rmtree(dst)
    except OSError as exc:
        if exc.errno != errno.ENOENT:
            raise CommandExecutionError(f"Unable to cleanup container root dir {dst}")
    raise CommandExecutionError(f"Container {name} failed to build")


def _bootstrap_arch(name, **kwargs):
    """
    Bootstrap an Arch Linux container
    """
    if not salt.utils.path.which("pacstrap"):
        raise CommandExecutionError(
            "pacstrap not found, is the arch-install-scripts package installed?"
        )
    dst = _make_container_root(name)
    cmd = f"pacstrap -c -d {dst} base"
    ret = __salt__["cmd.run_all"](cmd, python_shell=False)
    if ret["retcode"] != 0:
        _build_failed(dst, name)
    return ret


def _bootstrap_debian(name, **kwargs):
    """
    Bootstrap a Debian Linux container
    """
    version = kwargs.get("version", False)
    if not version:
        if __grains__["os"].lower() == "debian":
            version = __grains__["osrelease"]
        else:
            version = "stable"

    release_blacklist = [
        "hamm",
        "slink",
        "potato",
        "woody",
        "sarge",
        "etch",
        "lenny",
        "squeeze",
        "wheezy",
    ]
    if version in release_blacklist:
        raise CommandExecutionError(
            'Unsupported Debian version "{}". '
            'Only "stable" or "jessie" and newer are supported'.format(version)
        )

    dst = _make_container_root(name)
    cmd = f"debootstrap --arch=amd64 {version} {dst}"
    ret = __salt__["cmd.run_all"](cmd, python_shell=False)
    if ret["retcode"] != 0:
        _build_failed(dst, name)
    return ret


def _bootstrap_fedora(name, **kwargs):
    """
    Bootstrap a Fedora container
    """
    dst = _make_container_root(name)
    if not kwargs.get("version", False):
        if __grains__["os"].lower() == "fedora":
            version = __grains__["osrelease"]
        else:
            version = "21"
    else:
        version = "21"
    cmd = (
        "yum -y --releasever={} --nogpg --installroot={} "
        '--disablerepo="*" --enablerepo=fedora install systemd passwd yum '
        "fedora-release vim-minimal".format(version, dst)
    )
    ret = __salt__["cmd.run_all"](cmd, python_shell=False)
    if ret["retcode"] != 0:
        _build_failed(dst, name)
    return ret


def _bootstrap_ubuntu(name, **kwargs):
    """
    Bootstrap a Ubuntu Linux container
    """
    version = kwargs.get("version", False)
    if not version:
        if __grains__["os"].lower() == "ubuntu":
            version = __grains__["oscodename"]
        else:
            version = "xenial"
    dst = _make_container_root(name)
    cmd = f"debootstrap --arch=amd64 {version} {dst}"
    ret = __salt__["cmd.run_all"](cmd, python_shell=False)
    if ret["retcode"] != 0:
        _build_failed(dst, name)
    return ret


def _clear_context():
    """
    Clear any lxc variables set in __context__
    """
    for var in [x for x in __context__ if x.startswith("nspawn.")]:
        log.trace("Clearing __context__['%s']", var)
        __context__.pop(var, None)


def _ensure_running(name):
    """
    Raise an exception if the container does not exist
    """
    if state(name) != "running":
        return True
    else:
        return start(name)


def _ensure_systemd(version):
    """
    Raises an exception if the systemd version is not greater than the
    passed version.
    """
    try:
        version = int(version)
    except ValueError:
        raise CommandExecutionError(f"Invalid version '{version}'")

    try:
        installed = _sd_version()
        log.debug("nspawn: detected systemd %s", installed)
    except (IndexError, ValueError):
        raise CommandExecutionError("nspawn: Unable to get systemd version")

    if installed < version:
        raise CommandExecutionError(
            "This function requires systemd >= {} (Detected version: {}).".format(
                version, installed
            )
        )


def _machinectl(cmd, output_loglevel="debug", ignore_retcode=False, use_vt=False):
    """
    Helper function to run machinectl
    """
    prefix = "machinectl --no-legend --no-pager"
    return __salt__["cmd.run_all"](
        f"{prefix} {cmd}",
        output_loglevel=output_loglevel,
        ignore_retcode=ignore_retcode,
        use_vt=use_vt,
    )


@_ensure_exists
def _run(
    name,
    cmd,
    output=None,
    no_start=False,
    stdin=None,
    python_shell=True,
    preserve_state=False,
    output_loglevel="debug",
    ignore_retcode=False,
    use_vt=False,
    keep_env=None,
):
    """
    Common logic for nspawn.run functions
    """
    orig_state = state(name)
    exc = None
    try:
        ret = __salt__["container_resource.run"](
            name,
            cmd,
            container_type=__virtualname__,
            exec_driver=EXEC_DRIVER,
            output=output,
            no_start=no_start,
            stdin=stdin,
            python_shell=python_shell,
            output_loglevel=output_loglevel,
            ignore_retcode=ignore_retcode,
            use_vt=use_vt,
            keep_env=keep_env,
        )
    finally:
        # Make sure we stop the container if necessary, even if an exception
        # was raised.
        if preserve_state and orig_state == "stopped" and state(name) != "stopped":
            stop(name)

    if output in (None, "all"):
        return ret
    else:
        return ret[output]


@_ensure_exists
def pid(name):
    """
    Returns the PID of a container

    name
        Container name

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.pid arch1
    """
    try:
        return int(info(name).get("PID"))
    except (TypeError, ValueError) as exc:
        raise CommandExecutionError(f"Unable to get PID for container '{name}': {exc}")


def run(
    name,
    cmd,
    no_start=False,
    preserve_state=True,
    stdin=None,
    python_shell=True,
    output_loglevel="debug",
    use_vt=False,
    ignore_retcode=False,
    keep_env=None,
):
    """
    Run :mod:`cmd.run <salt.modules.cmdmod.run>` within a container

    name
        Name of the container in which to run the command

    cmd
        Command to run

    no_start : False
        If the container is not running, don't start it

    preserve_state : True
        After running the command, return the container to its previous state

    stdin : None
        Standard input to be used for the command

    output_loglevel : debug
        Level at which to log the output from the command. Set to ``quiet`` to
        suppress logging.

    use_vt : False
        Use SaltStack's utils.vt to stream output to console.

    keep_env : None
        If not passed, only a sane default PATH environment variable will be
        set. If ``True``, all environment variables from the container's host
        will be kept. Otherwise, a comma-separated list (or Python list) of
        environment variable names can be passed, and those environment
        variables will be kept.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.run mycontainer 'ifconfig -a'
    """
    return _run(
        name,
        cmd,
        output=None,
        no_start=no_start,
        preserve_state=preserve_state,
        stdin=stdin,
        python_shell=python_shell,
        output_loglevel=output_loglevel,
        use_vt=use_vt,
        ignore_retcode=ignore_retcode,
        keep_env=keep_env,
    )


def run_stdout(
    name,
    cmd,
    no_start=False,
    preserve_state=True,
    stdin=None,
    python_shell=True,
    output_loglevel="debug",
    use_vt=False,
    ignore_retcode=False,
    keep_env=None,
):
    """
    Run :mod:`cmd.run_stdout <salt.modules.cmdmod.run_stdout>` within a container

    name
        Name of the container in which to run the command

    cmd
        Command to run

    no_start : False
        If the container is not running, don't start it

    preserve_state : True
        After running the command, return the container to its previous state

    stdin : None
        Standard input to be used for the command

    output_loglevel : debug
        Level at which to log the output from the command. Set to ``quiet`` to
        suppress logging.

    use_vt : False
        Use SaltStack's utils.vt to stream output to console. Assumes
        ``output=all``.

    keep_env : None
        If not passed, only a sane default PATH environment variable will be
        set. If ``True``, all environment variables from the container's host
        will be kept. Otherwise, a comma-separated list (or Python list) of
        environment variable names can be passed, and those environment
        variables will be kept.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.run_stdout mycontainer 'ifconfig -a'
    """
    return _run(
        name,
        cmd,
        output="stdout",
        no_start=no_start,
        preserve_state=preserve_state,
        stdin=stdin,
        python_shell=python_shell,
        output_loglevel=output_loglevel,
        use_vt=use_vt,
        ignore_retcode=ignore_retcode,
        keep_env=keep_env,
    )


def run_stderr(
    name,
    cmd,
    no_start=False,
    preserve_state=True,
    stdin=None,
    python_shell=True,
    output_loglevel="debug",
    use_vt=False,
    ignore_retcode=False,
    keep_env=None,
):
    """
    Run :mod:`cmd.run_stderr <salt.modules.cmdmod.run_stderr>` within a container

    name
        Name of the container in which to run the command

    cmd
        Command to run

    no_start : False
        If the container is not running, don't start it

    preserve_state : True
        After running the command, return the container to its previous state

    stdin : None
        Standard input to be used for the command

    output_loglevel : debug
        Level at which to log the output from the command. Set to ``quiet`` to
        suppress logging.

    use_vt : False
        Use SaltStack's utils.vt to stream output to console. Assumes
        ``output=all``.

    keep_env : None
        If not passed, only a sane default PATH environment variable will be
        set. If ``True``, all environment variables from the container's host
        will be kept. Otherwise, a comma-separated list (or Python list) of
        environment variable names can be passed, and those environment
        variables will be kept.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.run_stderr mycontainer 'ip addr show'
    """
    return _run(
        name,
        cmd,
        output="stderr",
        no_start=no_start,
        preserve_state=preserve_state,
        stdin=stdin,
        python_shell=python_shell,
        output_loglevel=output_loglevel,
        use_vt=use_vt,
        ignore_retcode=ignore_retcode,
        keep_env=keep_env,
    )


def retcode(
    name,
    cmd,
    no_start=False,
    preserve_state=True,
    stdin=None,
    python_shell=True,
    output_loglevel="debug",
    use_vt=False,
    ignore_retcode=False,
    keep_env=None,
):
    """
    Run :mod:`cmd.retcode <salt.modules.cmdmod.retcode>` within a container

    name
        Name of the container in which to run the command

    cmd
        Command to run

    no_start : False
        If the container is not running, don't start it

    preserve_state : True
        After running the command, return the container to its previous state

    stdin : None
        Standard input to be used for the command

    output_loglevel : debug
        Level at which to log the output from the command. Set to ``quiet`` to
        suppress logging.

    use_vt : False
        Use SaltStack's utils.vt to stream output to console. Assumes
        ``output=all``.

    keep_env : None
        If not passed, only a sane default PATH environment variable will be
        set. If ``True``, all environment variables from the container's host
        will be kept. Otherwise, a comma-separated list (or Python list) of
        environment variable names can be passed, and those environment
        variables will be kept.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.retcode mycontainer 'ip addr show'
    """
    return _run(
        name,
        cmd,
        output="retcode",
        no_start=no_start,
        preserve_state=preserve_state,
        stdin=stdin,
        python_shell=python_shell,
        output_loglevel=output_loglevel,
        use_vt=use_vt,
        ignore_retcode=ignore_retcode,
        keep_env=keep_env,
    )


def run_all(
    name,
    cmd,
    no_start=False,
    preserve_state=True,
    stdin=None,
    python_shell=True,
    output_loglevel="debug",
    use_vt=False,
    ignore_retcode=False,
    keep_env=None,
):
    """
    Run :mod:`cmd.run_all <salt.modules.cmdmod.run_all>` within a container

    .. note::

        While the command is run within the container, it is initiated from the
        host. Therefore, the PID in the return dict is from the host, not from
        the container.

    name
        Name of the container in which to run the command

    cmd
        Command to run

    no_start : False
        If the container is not running, don't start it

    preserve_state : True
        After running the command, return the container to its previous state

    stdin : None
        Standard input to be used for the command

    output_loglevel : debug
        Level at which to log the output from the command. Set to ``quiet`` to
        suppress logging.

    use_vt : False
        Use SaltStack's utils.vt to stream output to console. Assumes
        ``output=all``.

    keep_env : None
        If not passed, only a sane default PATH environment variable will be
        set. If ``True``, all environment variables from the container's host
        will be kept. Otherwise, a comma-separated list (or Python list) of
        environment variable names can be passed, and those environment
        variables will be kept.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.run_all mycontainer 'ip addr show'
    """
    return _run(
        name,
        cmd,
        output="all",
        no_start=no_start,
        preserve_state=preserve_state,
        stdin=stdin,
        python_shell=python_shell,
        output_loglevel=output_loglevel,
        use_vt=use_vt,
        ignore_retcode=ignore_retcode,
        keep_env=keep_env,
    )


def bootstrap_container(name, dist=None, version=None):
    """
    Bootstrap a container from package servers, if dist is None the os the
    minion is running as will be created, otherwise the needed bootstrapping
    tools will need to be available on the host.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.bootstrap_container <name>
    """
    if not dist:
        dist = __grains__["os"].lower()
        log.debug("nspawn.bootstrap: no dist provided, defaulting to '%s'", dist)
    try:
        return globals()[f"_bootstrap_{dist}"](name, version=version)
    except KeyError:
        raise CommandExecutionError(f'Unsupported distribution "{dist}"')


def _needs_install(name):
    ret = 0
    has_minion = retcode(name, "command -v salt-minion")
    # we assume that installing is when no minion is running
    # but testing the executable presence is not enougth for custom
    # installs where the bootstrap can do much more than installing
    # the bare salt binaries.
    if has_minion:
        processes = run_stdout(name, "ps aux")
        if "salt-minion" not in processes:
            ret = 1
        else:
            retcode(name, "salt-call --local service.stop salt-minion")
    else:
        ret = 1
    return ret


def bootstrap_salt(
    name,
    config=None,
    approve_key=True,
    install=True,
    pub_key=None,
    priv_key=None,
    bootstrap_url=None,
    force_install=False,
    unconditional_install=False,
    bootstrap_delay=None,
    bootstrap_args=None,
    bootstrap_shell=None,
):
    """
    Bootstrap a container from package servers, if dist is None the os the
    minion is running as will be created, otherwise the needed bootstrapping
    tools will need to be available on the host.

    CLI Example:

    .. code-block:: bash

        salt '*' nspawn.bootstrap_salt arch1
    """
    if bootstrap_delay is not None:
        try:
            time.sleep(bootstrap_delay)
        except TypeError:
            # Bad input, but assume since a value was passed that
            # a delay was desired, and sleep for 5 seconds
            time.sleep(5)

    c_info = info(name)
    if not c_info:
        return None

    # default set here as we cannot set them
    # in def as it can come from a chain of procedures.
    if bootstrap_args:
        # custom bootstrap args can be totally customized, and user could
        # have inserted the placeholder for the config directory.
        # For example, some salt bootstrap script do not use at all -c
        if "{0}" not in bootstrap_args:
            bootstrap_args += " -c {0}"
    else:
        bootstrap_args = "-c {0}"
    if not bootstrap_shell:
        bootstrap_shell = "sh"

    orig_state = _ensure_running(name)
    if not orig_state:
        return orig_state
    if not force_install:
        needs_install = _needs_install(name)
    else:
        needs_install = True
    seeded = retcode(name, f"test -e '{SEED_MARKER}'") == 0
    tmp = tempfile.mkdtemp()
    if seeded and not unconditional_install:
        ret = True
    else:
        ret = False
        cfg_files = __salt__["seed.mkconfig"](
            config,
            tmp=tmp,
            id_=name,
            approve_key=approve_key,
            pub_key=pub_key,
            priv_key=priv_key,
        )
        if needs_install or force_install or unconditional_install:
            if install:
                rstr = __salt__["test.random_hash"]()
                configdir = f"/tmp/.c_{rstr}"
                run(name, f"install -m 0700 -d {configdir}", python_shell=False)
                bs_ = __salt__["config.gather_bootstrap_script"](
                    bootstrap=bootstrap_url
                )
                dest_dir = os.path.join("/tmp", rstr)
                for cmd in [
                    f"mkdir -p {dest_dir}",
                    f"chmod 700 {dest_dir}",
                ]:
                    if run_stdout(name, cmd):
                        log.error("tmpdir %s creation failed (%s)", dest_dir, cmd)
                        return False
                copy_to(name, bs_, f"{dest_dir}/bootstrap.sh", makedirs=True)
                copy_to(name, cfg_files["config"], os.path.join(configdir, "minion"))
                copy_to(
                    name, cfg_files["privkey"], os.path.join(configdir, "minion.pem")
                )
                copy_to(
                    name, cfg_files["pubkey"], os.path.join(configdir, "minion.pub")
                )
                bootstrap_args = bootstrap_args.format(configdir)
                cmd = "{0} {2}/bootstrap.sh {1}".format(
                    bootstrap_shell, bootstrap_args.replace("'", "''"), dest_dir
                )
                # log ASAP the forged bootstrap command which can be wrapped
                # out of the output in case of unexpected problem
                log.info("Running %s in LXC container '%s'", cmd, name)
                ret = retcode(name, cmd, output_loglevel="info", use_vt=True) == 0
            else:
                ret = False
        else:
            minion_config = salt.config.minion_config(cfg_files["config"])
            pki_dir = minion_config["pki_dir"]
            copy_to(name, cfg_files["config"], "/etc/salt/minion")
            copy_to(name, cfg_files["privkey"], os.path.join(pki_dir, "minion.pem"))
            copy_to(name, cfg_files["pubkey"], os.path.join(pki_dir, "minion.pub"))
            run(
                name, "salt-call --local service.enable salt-minion", python_shell=False
            )
            ret = True
        shutil.rmtree(tmp)
        if orig_state == "stopped":
            stop(name)
        # mark seeded upon successful install
        if ret:
            run(name, f"touch '{SEED_MARKER}'", python_shell=False)
    return ret


def list_all():
    """
    Lists all nspawn containers

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.list_all
    """
    ret = []
    if _sd_version() >= 219:
        for line in _machinectl("list-images")["stdout"].splitlines():
            try:
                ret.append(line.split()[0])
            except IndexError:
                continue
    else:
        rootdir = _root()
        try:
            for dirname in os.listdir(rootdir):
                if os.path.isdir(os.path.join(rootdir, dirname)):
                    ret.append(dirname)
        except OSError:
            pass
    return ret


def list_running():
    """
    Lists running nspawn containers

    .. note::

        ``nspawn.list`` also works to list running containers

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.list_running
        salt myminion nspawn.list
    """
    ret = []
    for line in _machinectl("list")["stdout"].splitlines():
        try:
            ret.append(line.split()[0])
        except IndexError:
            pass
    return sorted(ret)


# 'machinectl list' shows only running containers, so allow this to work as an
# alias to nspawn.list_running
list_ = salt.utils.functools.alias_function(list_running, "list_")


def list_stopped():
    """
    Lists stopped nspawn containers

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.list_stopped
    """
    return sorted(set(list_all()) - set(list_running()))


def exists(name):
    """
    Returns true if the named container exists

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.exists <name>
    """
    contextkey = f"nspawn.exists.{name}"
    if contextkey in __context__:
        return __context__[contextkey]
    __context__[contextkey] = name in list_all()
    return __context__[contextkey]


@_ensure_exists
def state(name):
    """
    Return state of container (running or stopped)

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.state <name>
    """
    try:
        cmd = f"show {name} --property=State"
        return _machinectl(cmd, ignore_retcode=True)["stdout"].split("=")[-1]
    except IndexError:
        return "stopped"


def info(name, **kwargs):
    """
    Return info about a container

    .. note::

        The container must be running for ``machinectl`` to gather information
        about it. If the container is stopped, then this function will start
        it.

    start : False
        If ``True``, then the container will be started to retrieve the info. A
        ``Started`` key will be in the return data if the container was
        started.

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.info arch1
        salt myminion nspawn.info arch1 force_start=False
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    start_ = kwargs.pop("start", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    if not start_:
        _ensure_running(name)
    elif name not in list_running():
        start(name)

    # Have to parse 'machinectl status' here since 'machinectl show' doesn't
    # contain IP address info or OS info. *shakes fist angrily*
    c_info = _machinectl(f"status {name}")
    if c_info["retcode"] != 0:
        raise CommandExecutionError(f"Unable to get info for container '{name}'")
    # Better human-readable names. False means key should be ignored.
    key_name_map = {
        "Iface": "Network Interface",
        "Leader": "PID",
        "Service": False,
        "Since": "Running Since",
    }
    ret = {}
    kv_pair = re.compile(r"^\s+([A-Za-z]+): (.+)$")
    tree = re.compile(r"[|`]")
    lines = c_info["stdout"].splitlines()
    multiline = False
    cur_key = None
    for idx, line in enumerate(lines):
        match = kv_pair.match(line)
        if match:
            key, val = match.groups()
            # Get a better key name if one exists
            key = key_name_map.get(key, key)
            if key is False:
                continue
            elif key == "PID":
                try:
                    val = val.split()[0]
                except IndexError:
                    pass
            cur_key = key
            if multiline:
                multiline = False
            ret[key] = val
        else:
            if cur_key is None:
                continue
            if tree.search(lines[idx]):
                # We've reached the process tree, bail out
                break
            if multiline:
                ret[cur_key].append(lines[idx].strip())
            else:
                ret[cur_key] = [ret[key], lines[idx].strip()]
                multiline = True
    return ret


@_ensure_exists
def enable(name):
    """
    Set the named container to be launched at boot

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.enable <name>
    """
    cmd = f"systemctl enable systemd-nspawn@{name}"
    if __salt__["cmd.retcode"](cmd, python_shell=False) != 0:
        __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
        return False
    return True


@_ensure_exists
def disable(name):
    """
    Set the named container to *not* be launched at boot

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.enable <name>
    """
    cmd = f"systemctl disable systemd-nspawn@{name}"
    if __salt__["cmd.retcode"](cmd, python_shell=False) != 0:
        __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
        return False
    return True


@_ensure_exists
def start(name):
    """
    Start the named container

    CLI Example:

    .. code-block:: bash

        salt myminion nspawn.start <name>
    """
    if _sd_version() >= 219:
        ret = _machinectl(f"start {name}")
    else:
        cmd = f"systemctl start systemd-nspawn@{name}"
        ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    if ret["retcode"] != 0:
        __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
        return False
    return True


# This function is hidden from sphinx docs
@_ensure_exists
def stop(name, kill=False):
    """
    This is a compatibility function which provides the logic for
    nspawn.poweroff and nspawn.terminate.
    """
    if _sd_version() >= 219:
        if kill:
            action = "terminate"
        else:
            action = "poweroff"
        ret = _machinectl(f"{action} {name}")
    else:
        cmd = f"systemctl stop systemd-nspawn@{name}"
        ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    if ret["retcode"] != 0:
        __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
        return False
    return True


def poweroff(name):
    """
    Issue a clean shutdown to the container.  Equivalent to running
    ``machinectl poweroff`` on the named container.

    For convenience, running ``nspawn.stop``(as shown in the CLI examples
    below) is equivalent to running ``nspawn.poweroff``.

    .. note::

        ``machinectl poweroff`` is only supported in systemd >= 219. On earlier
        systemd versions, running this function will simply issue a clean
        shutdown via ``systemctl``.

    CLI Examples:

    .. code-block:: bash

        salt myminion nspawn.poweroff arch1
        salt myminion nspawn.stop arch1
    """
    return stop(name, kill=False)


def terminate(name):
    """
    Kill all processes in the container without issuing a clean shutdown.
    Equivalent to running ``machinectl terminate`` on the named container.

    For convenience, running ``nspawn.stop`` and passing ``kill=True`` (as
    shown in the CLI examples below) is equivalent to running
    ``nspawn.terminate``.

    .. note::

        ``machinectl terminate`` is only supported in systemd >= 219. On
        earlier systemd versions, running this function will simply issue a
        clean shutdown via ``systemctl``.

    CLI Examples:

    .. code-block:: bash

        salt myminion nspawn.terminate arch1
        salt myminion nspawn.stop arch1 kill=True
    """
    return stop(name, kill=True)


# This function is hidden from sphinx docs
def restart(name):
    """
    This is a compatibility function which simply calls nspawn.reboot.
    """
    return reboot(name)


@_ensure_exists
def reboot(name, kill=False):
    """
    Reboot the container by sending a SIGINT to its init process. Equivalent
    to running ``machinectl reboot`` on the named container.

    For convenience, running ``nspawn.restart`` (as shown in the CLI examples
    below) is equivalent to running ``nspawn.reboot``.

    .. note::

        ``machinectl reboot`` is only supported in systemd >= 219. On earlier
        systemd versions, running this function will instead restart the
        container via ``systemctl``.

    CLI Examples:

    .. code-block:: bash

        salt myminion nspawn.reboot arch1
        salt myminion nspawn.restart arch1
    """
    if _sd_version() >= 219:
        if state(name) == "running":
            ret = _machinectl(f"reboot {name}")
        else:
            # 'machinectl reboot' will fail on a stopped container
            return start(name)
    else:
        # 'systemctl restart' did not work, at least in my testing. Running
        # 'uptime' in the container afterwards showed it had not rebooted. So,
        # we need stop and start the container in separate actions.

        # First stop the container
        cmd = f"systemctl stop systemd-nspawn@{name}"
        ret = __salt__["cmd.run_all"](cmd, python_shell=False)
        # Now check if successful
        if ret["retcode"] != 0:
            __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
            return False
        # Finally, start the container back up. No need to check the retcode a
        # second time, it'll be checked below once we exit the if/else block.
        cmd = f"systemctl start systemd-nspawn@{name}"
        ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    if ret["retcode"] != 0:
        __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
        return False
    return True


@_ensure_exists
def remove(name, stop=False):
    """
    Remove the named container

    .. warning::

        This function will remove all data associated with the container. It
        will not, however, remove the btrfs subvolumes created by pulling
        container images (:mod:`nspawn.pull_raw
        <salt.modules.nspawn.pull_raw>`, :mod:`nspawn.pull_tar
        <salt.modules.nspawn.pull_tar>`, :mod:`nspawn.pull_dkr
        <salt.modules.nspawn.pull_dkr>`).

    stop : False
        If ``True``, the container will be destroyed even if it is
        running/frozen.

    CLI Examples:

    .. code-block:: bash

        salt '*' nspawn.remove foo
        salt '*' nspawn.remove foo stop=True
    """
    if not stop and state(name) != "stopped":
        raise CommandExecutionError(f"Container '{name}' is not stopped")

    def _failed_remove(name, exc):
        raise CommandExecutionError(f"Unable to remove container '{name}': {exc}")

    if _sd_version() >= 219:
        ret = _machinectl(f"remove {name}")
        if ret["retcode"] != 0:
            __context__["retcode"] = salt.defaults.exitcodes.EX_UNAVAILABLE
            _failed_remove(name, ret["stderr"])
    else:
        try:
            shutil.rmtree(os.path.join(_root(), name))
        except OSError as exc:
            _failed_remove(name, exc)
    return True


# Compatibility between LXC and nspawn
destroy = salt.utils.functools.alias_function(remove, "destroy")


@_ensure_exists
def copy_to(name, source, dest, overwrite=False, makedirs=False):
    """
    Copy a file from the host into a container

    name
        Container name

    source
        File to be copied to the container

    dest
        Destination on the container. Must be an absolute path.

    overwrite : False
        Unless this option is set to ``True``, then if a file exists at the
        location specified by the ``dest`` argument, an error will be raised.

    makedirs : False

        Create the parent directory on the container if it does not already
        exist.

    CLI Example:

    .. code-block:: bash

        salt 'minion' nspawn.copy_to /tmp/foo /root/foo
    """
    path = source
    try:
        if source.startswith("salt://"):
            cached_source = __salt__["cp.cache_file"](source)
            if not cached_source:
                raise CommandExecutionError(f"Unable to cache {source}")
            path = cached_source
    except AttributeError:
        raise SaltInvocationError(f"Invalid source file {source}")

    if _sd_version() >= 219:
        # TODO: Use machinectl copy-to
        pass
    return __salt__["container_resource.copy_to"](
        name,
        path,
        dest,
        container_type=__virtualname__,
        exec_driver=EXEC_DRIVER,
        overwrite=overwrite,
        makedirs=makedirs,
    )


cp = salt.utils.functools.alias_function(copy_to, "cp")


# Everything below requres systemd >= 219
# TODO: Write a decorator to keep these functions from being available to older
#       systemd versions.
def _pull_image(pull_type, image, name, **kwargs):
    """
    Common logic for machinectl pull-* commands
    """
    _ensure_systemd(219)
    if exists(name):
        raise SaltInvocationError(f"Container '{name}' already exists")
    if pull_type in ("raw", "tar"):
        valid_kwargs = ("verify",)
    elif pull_type == "dkr":
        valid_kwargs = ("index",)
    else:
        raise SaltInvocationError(f"Unsupported image type '{pull_type}'")

    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    bad_kwargs = {
        x: y
        for x, y in salt.utils.args.clean_kwargs(**kwargs).items()
        if x not in valid_kwargs
    }

    if bad_kwargs:
        salt.utils.args.invalid_kwargs(bad_kwargs)

    pull_opts = []

    if pull_type in ("raw", "tar"):
        verify = kwargs.get("verify", False)
        if not verify:
            pull_opts.append("--verify=no")
        else:

            def _bad_verify():
                raise SaltInvocationError(
                    "'verify' must be one of the following: signature, checksum"
                )

            try:
                verify = verify.lower()
            except AttributeError:
                _bad_verify()
            else:
                if verify not in ("signature", "checksum"):
                    _bad_verify()
                pull_opts.append(f"--verify={verify}")

    elif pull_type == "dkr":
        # No need to validate the index URL, machinectl will take care of this
        # for us.
        if "index" in kwargs:
            pull_opts.append("--dkr-index-url={}".format(kwargs["index"]))

    cmd = "pull-{} {} {} {}".format(pull_type, " ".join(pull_opts), image, name)
    result = _machinectl(cmd, use_vt=True)
    if result["retcode"] != 0:
        msg = (
            "Error occurred pulling image. Stderr from the pull command "
            "(if any) follows: "
        )
        if result["stderr"]:
            msg += "\n\n{}".format(result["stderr"])
        raise CommandExecutionError(msg)
    return True


def pull_raw(url, name, verify=False):
    """
    Execute a ``machinectl pull-raw`` to download a .qcow2 or raw disk image,
    and add it to /var/lib/machines as a new container.

    .. note::

        **Requires systemd >= 219**

    url
        URL from which to download the container

    name
        Name for the new container

    verify : False
        Perform signature or checksum verification on the container. See the
        ``machinectl(1)`` man page (section titled "Image Transfer Commands")
        for more information on requirements for image verification. To perform
        signature verification, use ``verify=signature``. For checksum
        verification, use ``verify=checksum``. By default, no verification will
        be performed.

    CLI Examples:

    .. code-block:: bash

        salt myminion nspawn.pull_raw http://ftp.halifax.rwth-aachen.de/fedora/linux/releases/21/Cloud/Images/x86_64/Fedora-Cloud-Base-20141203-21.x86_64.raw.xz fedora21
    """
    return _pull_image("raw", url, name, verify=verify)


def pull_tar(url, name, verify=False):
    """
    Execute a ``machinectl pull-raw`` to download a .tar container image,
    and add it to /var/lib/machines as a new container.

    .. note::

        **Requires systemd >= 219**

    url
        URL from which to download the container

    name
        Name for the new container

    verify : False
        Perform signature or checksum verification on the container. See the
        ``machinectl(1)`` man page (section titled "Image Transfer Commands")
        for more information on requirements for image verification. To perform
        signature verification, use ``verify=signature``. For checksum
        verification, use ``verify=checksum``. By default, no verification will
        be performed.

    CLI Examples:

    .. code-block:: bash

        salt myminion nspawn.pull_tar http://foo.domain.tld/containers/archlinux-2015.02.01.tar.gz arch2
    """
    return _pull_image("tar", url, name, verify=verify)


def pull_dkr(url, name, index):
    """
    Execute a ``machinectl pull-dkr`` to download a docker image and add it to
    /var/lib/machines as a new container.

    .. note::

        **Requires systemd >= 219**

    url
        URL from which to download the container

    name
        Name for the new container

    index
        URL of the Docker index server from which to pull (must be an
        ``http://`` or ``https://`` URL).

    CLI Examples:

    .. code-block:: bash

        salt myminion nspawn.pull_dkr centos/centos6 cent6 index=https://get.docker.com
        salt myminion nspawn.pull_docker centos/centos6 cent6 index=https://get.docker.com
    """
    return _pull_image("dkr", url, name, index=index)


pull_docker = salt.utils.functools.alias_function(pull_dkr, "pull_docker")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.rh_service
"""

import textwrap

import pytest

import salt.modules.rh_service as rh_service
from tests.support.mock import MagicMock, patch


@pytest.fixture
def RET():
    return [
        "hostname",
        "mountall",
        "network-interface",
        "network-manager",
        "salt-api",
        "salt-master",
        "salt-minion",
    ]


@pytest.fixture
def configure_loader_modules():
    return {
        rh_service: {
            "_upstart_disable": None,
            "_upstart_enable": None,
            "_upstart_is_enabled": None,
        }
    }


def _m_lst():
    """
    Return value for [].
    """
    return MagicMock(return_value=[])


def _m_ret(RET):
    """
    Return value for RET.
    """
    return MagicMock(return_value=RET)


def _m_bool(bol=True):
    """
    Return Bool value.
    """
    return MagicMock(return_value=bol)


def test__chkconfig_is_enabled():
    """
    test _chkconfig_is_enabled function
    """
    name = "atd"
    chkconfig_out = textwrap.dedent(
        """\

        {}           0:off   1:off   2:off   3:on    4:on    5:on    6:off
        """.format(
            name
        )
    )
    xinetd_out = textwrap.dedent(
        """\
        xinetd based services:
                {}  on
        """.format(
            name
        )
    )

    with patch.object(rh_service, "_runlevel", MagicMock(return_value=3)):
        mock_run = MagicMock(return_value={"retcode": 0, "stdout": chkconfig_out})
        with patch.dict(rh_service.__salt__, {"cmd.run_all": mock_run}):
            assert rh_service._chkconfig_is_enabled(name)
            assert not rh_service._chkconfig_is_enabled(name, 2)
            assert rh_service._chkconfig_is_enabled(name, 3)

        mock_run = MagicMock(return_value={"retcode": 0, "stdout": xinetd_out})
        with patch.dict(rh_service.__salt__, {"cmd.run_all": mock_run}):
            assert rh_service._chkconfig_is_enabled(name)
            assert rh_service._chkconfig_is_enabled(name, 2)
            assert rh_service._chkconfig_is_enabled(name, 3)


# 'get_enabled' function tests: 1


def test_get_enabled(RET):
    """
    Test if it return the enabled services. Use the ``limit``
    param to restrict results to services of that type.
    """
    with patch.object(rh_service, "_upstart_services", _m_ret(RET)):
        with patch.object(
            rh_service, "_upstart_is_enabled", MagicMock(return_value=False)
        ):
            assert rh_service.get_enabled("upstart") == []

    mock_run = MagicMock(return_value="salt stack")
    with patch.dict(rh_service.__salt__, {"cmd.run": mock_run}):
        with patch.object(rh_service, "_sysv_services", _m_ret(RET)):
            with patch.object(rh_service, "_sysv_is_enabled", _m_bool()):
                assert rh_service.get_enabled("sysvinit") == RET

                with patch.object(rh_service, "_upstart_services", _m_lst()):
                    with patch.object(
                        rh_service,
                        "_upstart_is_enabled",
                        MagicMock(return_value=True),
                    ):
                        assert rh_service.get_enabled() == RET


# 'get_disabled' function tests: 1


def test_get_disabled(RET):
    """
    Test if it return the disabled services. Use the ``limit``
    param to restrict results to services of that type.
    """
    with patch.object(rh_service, "_upstart_services", _m_ret(RET)):
        with patch.object(
            rh_service, "_upstart_is_enabled", MagicMock(return_value=False)
        ):
            assert rh_service.get_disabled("upstart") == RET

    mock_run = MagicMock(return_value="salt stack")
    with patch.dict(rh_service.__salt__, {"cmd.run": mock_run}):
        with patch.object(rh_service, "_sysv_services", _m_ret(RET)):
            with patch.object(rh_service, "_sysv_is_enabled", _m_bool(False)):
                assert rh_service.get_disabled("sysvinit") == RET

                with patch.object(rh_service, "_upstart_services", _m_lst()):
                    with patch.object(
                        rh_service,
                        "_upstart_is_enabled",
                        MagicMock(return_value=False),
                    ):
                        assert rh_service.get_disabled() == RET


# 'get_all' function tests: 1


def test_get_all(RET):
    """
    Test if it return all installed services. Use the ``limit``
    param to restrict results to services of that type.
    """
    with patch.object(rh_service, "_upstart_services", _m_ret(RET)):
        assert rh_service.get_all("upstart") == RET

    with patch.object(rh_service, "_sysv_services", _m_ret(RET)):
        assert rh_service.get_all("sysvinit") == RET

        with patch.object(rh_service, "_upstart_services", _m_lst()):
            assert rh_service.get_all() == RET


# 'available' function tests: 1


def test_available():
    """
    Test if it return True if the named service is available.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool()):
        assert rh_service.available("salt-api", "upstart")

    with patch.object(rh_service, "_service_is_sysv", _m_bool()):
        assert rh_service.available("salt-api", "sysvinit")

        with patch.object(rh_service, "_service_is_upstart", _m_bool()):
            assert rh_service.available("salt-api")


# 'missing' function tests: 1


def test_missing():
    """
    Test if it return True if the named service is not available.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool(False)):
        assert rh_service.missing("sshd", "upstart")

        with patch.object(rh_service, "_service_is_sysv", _m_bool(False)):
            assert rh_service.missing("sshd")

    with patch.object(rh_service, "_service_is_sysv", _m_bool()):
        assert not rh_service.missing("sshd", "sysvinit")

        with patch.object(rh_service, "_service_is_upstart", _m_bool()):
            assert not rh_service.missing("sshd")


# 'start' function tests: 1


def test_start():
    """
    Test if it start the specified service.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool()):
        with patch.dict(rh_service.__salt__, {"cmd.retcode": _m_bool(False)}):
            assert rh_service.start("salt-api")


# 'stop' function tests: 1


def test_stop():
    """
    Test if it stop the specified service.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool()):
        with patch.dict(rh_service.__salt__, {"cmd.retcode": _m_bool(False)}):
            assert rh_service.stop("salt-api")


# 'restart' function tests: 1


def test_restart():
    """
    Test if it restart the specified service.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool()):
        with patch.dict(rh_service.__salt__, {"cmd.retcode": _m_bool(False)}):
            assert rh_service.restart("salt-api")


# 'reload_' function tests: 1


def test_reload():
    """
    Test if it reload the specified service.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool()):
        with patch.dict(rh_service.__salt__, {"cmd.retcode": _m_bool(False)}):
            assert rh_service.reload_("salt-api")


# 'status' function tests: 1


def test_status():
    """
    Test if it return the status for a service,
    returns a bool whether the service is running.
    """
    with patch.object(rh_service, "_service_is_upstart", _m_bool()):
        mock_run = MagicMock(return_value="start/running")
        with patch.dict(rh_service.__salt__, {"cmd.run": mock_run}):
            assert rh_service.status("salt-api")

    with patch.object(rh_service, "_service_is_upstart", _m_bool(False)):
        with patch.dict(rh_service.__salt__, {"status.pid": _m_bool()}):
            assert rh_service.status("salt-api", sig=True)

        mock_ret = MagicMock(return_value=0)
        with patch.dict(rh_service.__salt__, {"cmd.retcode": mock_ret}):
            assert rh_service.status("salt-api")


# 'enable' function tests: 1


def test_enable():
    """
    Test if it enable the named service to start at boot.
    """
    mock_bool = MagicMock(side_effect=[True, True, False])
    with patch.object(rh_service, "_service_is_upstart", mock_bool):
        with patch.object(
            rh_service, "_upstart_is_enabled", MagicMock(return_value=True)
        ):
            with patch.object(
                rh_service, "_upstart_enable", MagicMock(return_value=False)
            ):
                assert not rh_service.enable("salt-api")
            with patch.object(
                rh_service, "_upstart_enable", MagicMock(return_value=True)
            ):
                assert rh_service.enable("salt-api")

        with patch.object(rh_service, "_sysv_enable", _m_bool()):
            assert rh_service.enable("salt-api")


# 'disable' function tests: 1


def test_disable():
    """
    Test if it disable the named service to start at boot.
    """
    mock_bool = MagicMock(side_effect=[True, True, False])
    with patch.object(rh_service, "_service_is_upstart", mock_bool):
        with patch.object(
            rh_service, "_upstart_is_enabled", MagicMock(return_value=True)
        ):
            with patch.object(
                rh_service, "_upstart_disable", MagicMock(return_value=False)
            ):
                assert not rh_service.disable("salt-api")
            with patch.object(
                rh_service, "_upstart_disable", MagicMock(return_value=True)
            ):
                assert rh_service.disable("salt-api")

        with patch.object(rh_service, "_sysv_disable", _m_bool()):
            assert rh_service.disable("salt-api")


# 'enabled' function tests: 1


def test_enabled():
    """
    Test if it check to see if the named service is enabled
    to start on boot.
    """
    mock_bool = MagicMock(side_effect=[True, False])
    with patch.object(rh_service, "_service_is_upstart", mock_bool):
        with patch.object(
            rh_service, "_upstart_is_enabled", MagicMock(return_value=False)
        ):
            assert not rh_service.enabled("salt-api")

        with patch.object(rh_service, "_sysv_is_enabled", _m_bool()):
            assert rh_service.enabled("salt-api")


# 'disabled' function tests: 1


def test_disabled():
    """
    Test if it check to see if the named service is disabled
    to start on boot.
    """
    mock_bool = MagicMock(side_effect=[True, False])
    with patch.object(rh_service, "_service_is_upstart", mock_bool):
        with patch.object(
            rh_service, "_upstart_is_enabled", MagicMock(return_value=False)
        ):
            assert rh_service.disabled("salt-api")

        with patch.object(rh_service, "_sysv_is_enabled", _m_bool(False)):
            assert rh_service.disabled("salt-api")



"""
Create and verify ANSI X9.31 RSA signatures using OpenSSL libcrypto
"""

import ctypes.util
import glob
import os
import platform
import sys
from ctypes import c_char_p, c_int, c_void_p, cdll, create_string_buffer, pointer

import salt.utils.platform
import salt.utils.stringutils

# Constants taken from openssl-1.1.0c/include/openssl/crypto.h
OPENSSL_INIT_ADD_ALL_CIPHERS = 0x00000004
OPENSSL_INIT_ADD_ALL_DIGESTS = 0x00000008
OPENSSL_INIT_NO_LOAD_CONFIG = 0x00000080


def _find_libcrypto():
    """
    Find the path (or return the short name) of libcrypto.
    """
    if sys.platform.startswith("win"):
        lib = None
        for path in sys.path:
            lib = glob.glob(os.path.join(path, "libcrypto*.dll"))
            lib = lib[0] if lib else None
            if lib:
                break

    elif salt.utils.platform.is_darwin():
        # will look for several different location on the system,
        # Search in the following order:
        # - salt's pkg install location
        # - relative to the running python (sys.executable)
        # - homebrew
        # - macports
        # - system libraries

        # look in salts pkg install location.
        lib = glob.glob("/opt/salt/lib/libcrypto.dylib")

        # look in location salt is running from
        # this accounts for running from an unpacked onedir file
        lib = lib or glob.glob("lib/libcrypto.dylib")

        # Look in the location relative to the python binary
        # Try to account for this being a venv by resolving the path if it is a
        # symlink
        py_bin = sys.executable
        if os.path.islink(py_bin):
            py_bin = os.path.realpath(py_bin)
        target = os.path.dirname(py_bin)
        if os.path.basename(target) == "bin":
            target = os.path.dirname(target)
        lib = lib or glob.glob(f"{target}/lib/libcrypto.dylib")

        # Find library symlinks in Homebrew locations.
        brew_prefix = os.getenv("HOMEBREW_PREFIX", "/usr/local")
        lib = lib or glob.glob(
            os.path.join(brew_prefix, "opt/openssl/lib/libcrypto.dylib")
        )
        lib = lib or glob.glob(
            os.path.join(brew_prefix, "opt/openssl@*/lib/libcrypto.dylib")
        )
        # look in macports.
        lib = lib or glob.glob("/opt/local/lib/libcrypto.dylib")
        # check if 10.15, regular libcrypto.dylib is just a false pointer.
        if platform.mac_ver()[0].split(".")[:2] == ["10", "15"]:
            lib = lib or glob.glob("/usr/lib/libcrypto.*.dylib")
            lib = list(reversed(sorted(lib)))
        elif int(platform.mac_ver()[0].split(".")[0]) < 11:
            # Fall back on system libcrypto (only works before Big Sur)
            lib = lib or ["/usr/lib/libcrypto.dylib"]
        lib = lib[0] if lib else None
    elif getattr(sys, "frozen", False) and salt.utils.platform.is_smartos():
        lib = glob.glob(os.path.join(os.path.dirname(sys.executable), "libcrypto.so*"))
        lib = lib[0] if lib else None
    else:
        lib = ctypes.util.find_library("crypto")
        if not lib:
            if salt.utils.platform.is_sunos():
                # Solaris-like distribution that use pkgsrc have libraries
                # in a non standard location.
                # (SmartOS, OmniOS, OpenIndiana, ...)
                # This could be /opt/tools/lib (Global Zone) or
                # /opt/local/lib (non-Global Zone), thus the two checks
                # below
                lib = glob.glob("/opt/saltstack/salt/run/libcrypto.so*")
                lib = lib or glob.glob("/opt/local/lib/libcrypto.so*")
                lib = lib or glob.glob("/opt/tools/lib/libcrypto.so*")
                lib = lib[0] if lib else None
            elif salt.utils.platform.is_aix():
                if os.path.isdir("/opt/saltstack/salt/run") or os.path.isdir(
                    "/opt/salt/lib"
                ):
                    # preference for Salt installed fileset
                    lib = glob.glob("/opt/saltstack/salt/run/libcrypto.so*")
                    lib = lib or glob.glob("/opt/salt/lib/libcrypto.so*")
                else:
                    lib = glob.glob("/opt/freeware/lib/libcrypto.so*")
                lib = lib[0] if lib else None
    if not lib:
        raise OSError("Cannot locate OpenSSL libcrypto")
    return lib


def _load_libcrypto():
    """
    Attempt to load libcrypto.
    """
    return cdll.LoadLibrary(_find_libcrypto())


def _init_libcrypto():
    """
    Set up libcrypto argtypes and initialize the library
    """
    libcrypto = _load_libcrypto()

    try:
        # If we're greater than OpenSSL 1.1.0, no need to to the init
        openssl_version_num = libcrypto.OpenSSL_version_num
        if callable(openssl_version_num):
            openssl_version_num = openssl_version_num()
        if openssl_version_num < 0x10100000:
            libcrypto.OPENSSL_init_crypto()
    except AttributeError:
        # Support for OpenSSL < 1.1 (OPENSSL_API_COMPAT < 0x10100000L)
        libcrypto.OPENSSL_no_config()
        libcrypto.OPENSSL_add_all_algorithms_noconf()

    libcrypto.RSA_new.argtypes = ()
    libcrypto.RSA_new.restype = c_void_p
    libcrypto.RSA_free.argtypes = (c_void_p,)
    libcrypto.RSA_size.argtype = c_void_p
    libcrypto.BIO_new_mem_buf.argtypes = (c_char_p, c_int)
    libcrypto.BIO_new_mem_buf.restype = c_void_p
    libcrypto.BIO_free.argtypes = (c_void_p,)
    libcrypto.PEM_read_bio_RSAPrivateKey.argtypes = (
        c_void_p,
        c_void_p,
        c_void_p,
        c_void_p,
    )
    libcrypto.PEM_read_bio_RSAPrivateKey.restype = c_void_p
    libcrypto.PEM_read_bio_RSA_PUBKEY.argtypes = (
        c_void_p,
        c_void_p,
        c_void_p,
        c_void_p,
    )
    libcrypto.PEM_read_bio_RSA_PUBKEY.restype = c_void_p
    libcrypto.RSA_private_encrypt.argtypes = (
        c_int,
        c_char_p,
        c_char_p,
        c_void_p,
        c_int,
    )
    libcrypto.RSA_public_decrypt.argtypes = (c_int, c_char_p, c_char_p, c_void_p, c_int)

    return libcrypto


libcrypto = _init_libcrypto()

# openssl/rsa.h:#define RSA_X931_PADDING 5
RSA_X931_PADDING = 5


class RSAX931Signer:
    """
    Create ANSI X9.31 RSA signatures using OpenSSL libcrypto
    """

    def __init__(self, keydata):
        """
        Init an RSAX931Signer instance

        :param str keydata: The RSA private key in PEM format
        """
        keydata = salt.utils.stringutils.to_bytes(keydata, "ascii")
        self._bio = libcrypto.BIO_new_mem_buf(keydata, len(keydata))
        self._rsa = c_void_p(libcrypto.RSA_new())
        if not libcrypto.PEM_read_bio_RSAPrivateKey(
            self._bio, pointer(self._rsa), None, None
        ):
            raise ValueError("invalid RSA private key")

    # pylint: disable=W1701
    def __del__(self):
        libcrypto.BIO_free(self._bio)
        libcrypto.RSA_free(self._rsa)

    # pylint: enable=W1701

    def sign(self, msg):
        """
        Sign a message (digest) using the private key

        :param str msg: The message (digest) to sign
        :rtype: str
        :return: The signature, or an empty string if the encryption failed
        """
        # Allocate a buffer large enough for the signature. Freed by ctypes.
        buf = create_string_buffer(libcrypto.RSA_size(self._rsa))
        msg = salt.utils.stringutils.to_bytes(msg)
        size = libcrypto.RSA_private_encrypt(
            len(msg), msg, buf, self._rsa, RSA_X931_PADDING
        )
        if size < 0:
            raise ValueError("Unable to encrypt message")
        return buf[0:size]


class RSAX931Verifier:
    """
    Verify ANSI X9.31 RSA signatures using OpenSSL libcrypto
    """

    def __init__(self, pubdata):
        """
        Init an RSAX931Verifier instance

        :param str pubdata: The RSA public key in PEM format
        """
        pubdata = salt.utils.stringutils.to_bytes(pubdata, "ascii")
        pubdata = pubdata.replace(b"RSA ", b"")
        self._bio = libcrypto.BIO_new_mem_buf(pubdata, len(pubdata))
        self._rsa = c_void_p(libcrypto.RSA_new())
        if not libcrypto.PEM_read_bio_RSA_PUBKEY(
            self._bio, pointer(self._rsa), None, None
        ):
            raise ValueError("invalid RSA public key")

    # pylint: disable=W1701
    def __del__(self):
        libcrypto.BIO_free(self._bio)
        libcrypto.RSA_free(self._rsa)

    # pylint: enable=W1701

    def verify(self, signed):
        """
        Recover the message (digest) from the signature using the public key

        :param str signed: The signature created with the private key
        :rtype: str
        :return: The message (digest) recovered from the signature, or an empty
            string if the decryption failed
        """
        # Allocate a buffer large enough for the signature. Freed by ctypes.
        buf = create_string_buffer(libcrypto.RSA_size(self._rsa))
        signed = salt.utils.stringutils.to_bytes(signed)
        size = libcrypto.RSA_public_decrypt(
            len(signed), signed, buf, self._rsa, RSA_X931_PADDING
        )
        if size < 0:
            raise ValueError("Unable to decrypt message")
        return buf[0:size]



import os
import subprocess

import pytest

import salt.utils.path
import salt.utils.platform
import salt.utils.systemd
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


def _check_systemctl():
    if not hasattr(_check_systemctl, "memo"):
        if not salt.utils.platform.is_linux():
            _check_systemctl.memo = False
        else:
            proc = subprocess.run(["systemctl"], capture_output=True, check=False)
            _check_systemctl.memo = (
                b"Failed to get D-Bus connection: No such file or directory"
                in proc.stderr
            )
    return _check_systemctl.memo


@pytest.fixture
def service_name(grains, modules):
    # For local testing purposes
    env_name = os.environ.get("SALT_FUNCTIONAL_TEST_SERVICE_NAME")
    if env_name is not None:
        return env_name

    service_name = "cron"
    cmd_name = "crontab"
    os_family = grains.get("os_family")
    is_systemd = grains.get("systemd")
    if os_family == "RedHat":
        service_name = "crond"
    elif os_family == "Arch":
        service_name = "sshd"
        cmd_name = "systemctl"
    elif os_family == "MacOS":
        service_name = "com.apple.AirPlayXPCHelper"
    elif os_family == "Windows":
        service_name = "Spooler"

    if os_family != "Windows" and salt.utils.path.which(cmd_name) is None:
        pytest.skip(f"{cmd_name} is not installed")

    if is_systemd and modules.service.offline():
        pytest.skip("systemd is OFFLINE")

    return service_name


@pytest.fixture(autouse=True)
def setup_service(service_name, modules):
    pre_srv_status = modules.service.status(service_name)
    pre_srv_enabled = service_name in modules.service.get_enabled()

    try:
        yield pre_srv_status
    finally:
        post_srv_status = modules.service.status(service_name)
        post_srv_enabled = service_name in modules.service.get_enabled()

        if post_srv_status != pre_srv_status:
            if pre_srv_status:
                modules.service.start(service_name)
            else:
                modules.service.stop(service_name)

        if post_srv_enabled != pre_srv_enabled:
            if pre_srv_enabled:
                modules.service.enable(service_name)
            else:
                modules.service.disable(service_name)


@pytest.mark.skipif(_check_systemctl(), reason="systemctl degraded")
def test_service_status_running(modules, service_name):
    """
    test service.status execution module
    when service is running
    """
    modules.service.start(service_name)
    check_service = modules.service.status(service_name)
    assert check_service


def test_service_status_dead(modules, service_name):
    """
    test service.status execution module
    when service is dead
    """
    modules.service.stop(service_name)
    check_service = modules.service.status(service_name)
    assert not check_service


@pytest.mark.skipif(_check_systemctl(), reason="systemctl degraded")
def test_service_restart(modules, service_name):
    """
    test service.restart
    """
    assert modules.service.stop(service_name)


@pytest.mark.skipif(_check_systemctl(), reason="systemctl degraded")
def test_service_enable(modules, service_name):
    """
    test service.get_enabled and service.enable module
    """
    # disable service before test
    assert modules.service.disable(service_name)

    assert modules.service.enable(service_name)
    assert service_name in modules.service.get_enabled()


@pytest.mark.skipif(_check_systemctl(), reason="systemctl degraded")
def test_service_disable(modules, service_name):
    """
    test service.get_disabled and service.disable module
    """
    # enable service before test
    assert modules.service.enable(service_name)

    assert modules.service.disable(service_name)
    if salt.utils.platform.is_darwin():
        assert modules.service.disabled(service_name)
    else:
        assert service_name in modules.service.get_disabled()


def test_service_disable_doesnot_exist(modules):
    """
    test service.get_disabled and service.disable module
    when service name does not exist
    """
    # enable service before test
    srv_name = "doesnotexist"
    try:
        enable = modules.service.enable(srv_name)
        assert not enable
    except CommandExecutionError as exc:
        assert srv_name in exc.error or "no such file or directory" in exc.error.lower()

    try:
        disable = modules.service.disable(srv_name)
        assert not disable
    except CommandExecutionError as exc:
        assert srv_name in exc.error or "no such file or directory" in exc.error.lower()

    if salt.utils.platform.is_darwin():
        with pytest.raises(
            CommandExecutionError, match=f"Service not found: {srv_name}"
        ):
            modules.service.disabled(srv_name)
    else:
        assert srv_name not in modules.service.get_disabled()


@pytest.mark.skip_unless_on_windows
def test_service_get_service_name(modules, service_name):
    """
    test service.get_service_name
    """
    ret = modules.service.get_service_name()
    assert service_name in ret.values()



"""
Saltutil State
==============

This state wraps the saltutil execution modules to make them easier to run
from a states. Rather than needing to to use ``module.run`` this state allows for
improved change detection.

    .. versionadded:: 3000
"""

import logging

# Define the module's virtual name
__virtualname__ = "saltutil"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Named saltutil
    """
    return __virtualname__


def _sync_single(name, module, **kwargs):
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"saltutil.sync_{module} would have been run"
        return ret

    try:
        sync_status = __salt__[f"saltutil.sync_{module}"](**kwargs)
        if sync_status:
            ret["changes"][module] = sync_status
            ret["comment"] = f"Updated {module}."
    except Exception as e:  # pylint: disable=broad-except
        log.error("Failed to run saltutil.sync_%s: %s", module, e)
        ret["result"] = False
        ret["comment"] = f"Failed to run sync_{module}: {e}"
        return ret

    if not ret["changes"]:
        ret["comment"] = "No updates to sync"

    return ret


def sync_all(name, **kwargs):
    """
    Performs the same task as saltutil.sync_all module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_all:
            - refresh: True
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "saltutil.sync_all would have been run"
        return ret

    try:
        sync_status = __salt__["saltutil.sync_all"](**kwargs)
        for key, value in sync_status.items():
            if value:
                ret["changes"][key] = value
                ret["comment"] = "Sync performed"
    except Exception as e:  # pylint: disable=broad-except
        log.error("Failed to run saltutil.sync_all: %s", e)
        ret["result"] = False
        ret["comment"] = f"Failed to run sync_all: {e}"
        return ret

    if not ret["changes"]:
        ret["comment"] = "No updates to sync"

    return ret


def sync_beacons(name, **kwargs):
    """
    Performs the same task as saltutil.sync_beacons module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_beacons:
            - refresh: True
    """
    return _sync_single(name, "beacons", **kwargs)


def sync_clouds(name, **kwargs):
    """
    Performs the same task as saltutil.sync_clouds module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_clouds:
            - refresh: True
    """
    return _sync_single(name, "clouds", **kwargs)


def sync_engines(name, **kwargs):
    """
    Performs the same task as saltutil.sync_engines module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_engines:
            - refresh: True
    """
    return _sync_single(name, "engines", **kwargs)


def sync_grains(name, **kwargs):
    """
    Performs the same task as saltutil.sync_grains module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_grains:
            - refresh: True
    """
    return _sync_single(name, "grains", **kwargs)


def sync_executors(name, **kwargs):
    """
    Performs the same task as saltutil.sync_executors module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_executors:
            - refresh: True
    """
    return _sync_single(name, "executors", **kwargs)


def sync_log_handlers(name, **kwargs):
    """
    Performs the same task as saltutil.sync_log_handlers module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_log_handlers:
            - refresh: True
    """
    return _sync_single(name, "log_handlers", **kwargs)


def sync_modules(name, **kwargs):
    """
    Performs the same task as saltutil.sync_modules module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_modules:
            - refresh: True
    """
    return _sync_single(name, "modules", **kwargs)


def sync_output(name, **kwargs):
    """
    Performs the same task as saltutil.sync_output module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_output:
            - refresh: True
    """
    return _sync_single(name, "output", **kwargs)


def sync_outputters(name, **kwargs):
    """
    Performs the same task as saltutil.sync_outputters module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_outputters:
            - refresh: True
    """
    return _sync_single(name, "outputters", **kwargs)


def sync_pillar(name, **kwargs):
    """
    Performs the same task as saltutil.sync_pillar module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_pillar:
            - refresh: True
    """
    return _sync_single(name, "pillar", **kwargs)


def sync_proxymodules(name, **kwargs):
    """
    Performs the same task as saltutil.sync_proxymodules module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_proxymodules:
            - refresh: True
    """
    return _sync_single(name, "proxymodules", **kwargs)


def sync_matchers(name, **kwargs):
    """
    Performs the same task as saltutil.sync_matchers module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_matchers:
            - refresh: True
    """
    return _sync_single(name, "matchers", **kwargs)


def sync_renderers(name, **kwargs):
    """
    Performs the same task as saltutil.sync_renderers module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_renderers:
            - refresh: True
    """
    return _sync_single(name, "renderers", **kwargs)


def sync_returners(name, **kwargs):
    """
    Performs the same task as saltutil.sync_returners module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_returners:
            - refresh: True
    """
    return _sync_single(name, "returners", **kwargs)


def sync_sdb(name, **kwargs):
    """
    Performs the same task as saltutil.sync_sdb module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_sdb:
            - refresh: True
    """
    return _sync_single(name, "sdb", **kwargs)


def sync_states(name, **kwargs):
    """
    Performs the same task as saltutil.sync_states module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_states:
            - refresh: True
    """
    return _sync_single(name, "states", **kwargs)


def sync_thorium(name, **kwargs):
    """
    Performs the same task as saltutil.sync_thorium module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_thorium:
            - refresh: True
    """
    return _sync_single(name, "thorium", **kwargs)


def sync_utils(name, **kwargs):
    """
    Performs the same task as saltutil.sync_utils module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_utils:
            - refresh: True
    """
    return _sync_single(name, "utils", **kwargs)


def sync_serializers(name, **kwargs):
    """
    Performs the same task as saltutil.sync_serializers module
    See :mod:`saltutil module for full list of options <salt.modules.saltutil>`

    .. code-block:: yaml

        sync_everything:
          saltutil.sync_serializers:
            - refresh: True
    """
    return _sync_single(name, "serializers", **kwargs)



"""
Tests for salt.utils.data
"""

import builtins
import logging
from collections import OrderedDict as SaltOrderedDict

import pytest

import salt.utils.data
import salt.utils.stringutils
from tests.support.mock import patch
from tests.support.unit import LOREM_IPSUM

log = logging.getLogger(__name__)


def _b(x):
    return x.encode("utf-8")


def _s(x):
    return salt.utils.stringutils.to_str(x, normalize=True)


@pytest.fixture
def get_BYTES():
    # Some randomized data that will not decode
    return b"1\x814\x10"


@pytest.fixture
def get_EGGS():
    # This is an example of a unicode string with й constructed using two separate
    # code points. Do not modify it.
    return "\u044f\u0438\u0306\u0446\u0430"


@pytest.fixture
def get_test_data(get_BYTES, get_EGGS):
    return [
        "unicode_str",
        _b("питон"),
        123,
        456.789,
        True,
        False,
        None,
        get_EGGS,
        get_BYTES,
        [123, 456.789, _b("спам"), True, False, None, get_EGGS, get_BYTES],
        (987, 654.321, _b("яйца"), get_EGGS, None, (True, get_EGGS, get_BYTES)),
        {
            _b("str_key"): _b("str_val"),
            None: True,
            123: 456.789,
            get_EGGS: get_BYTES,
            _b("subdict"): {
                "unicode_key": get_EGGS,
                _b("tuple"): (123, "hello", _b("world"), True, get_EGGS, get_BYTES),
                _b("list"): [456, _b("спам"), False, get_EGGS, get_BYTES],
            },
        },
        SaltOrderedDict([(_b("foo"), "bar"), (123, 456), (get_EGGS, get_BYTES)]),
    ]


def test_sorted_ignorecase():
    test_list = ["foo", "Foo", "bar", "Bar"]
    expected_list = ["bar", "Bar", "foo", "Foo"]
    assert salt.utils.data.sorted_ignorecase(test_list) == expected_list


def test_mysql_to_dict():
    test_mysql_output = [
        "+----+------+-----------+------+---------+------+-------+------------------+",
        "| Id | User | Host      | db   | Command | Time | State | Info         "
        "    |",
        "+----+------+-----------+------+---------+------+-------+------------------+",
        "|  7 | root | localhost | NULL | Query   |    0 | init  | show"
        " processlist |",
        "+----+------+-----------+------+---------+------+-------+------------------+",
    ]

    ret = salt.utils.data.mysql_to_dict(test_mysql_output, "Info")
    expected_dict = {
        "show processlist": {
            "Info": "show processlist",
            "db": "NULL",
            "State": "init",
            "Host": "localhost",
            "Command": "Query",
            "User": "root",
            "Time": 0,
            "Id": 7,
        }
    }
    assert ret == expected_dict


def test_subdict_match():
    test_two_level_dict = {"foo": {"bar": "baz"}}
    test_two_level_comb_dict = {"foo": {"bar": "baz:woz"}}
    test_two_level_dict_and_list = {
        "abc": ["def", "ghi", {"lorem": {"ipsum": [{"dolor": "sit"}]}}],
    }
    test_three_level_dict = {"a": {"b": {"c": "v"}}}

    assert salt.utils.data.subdict_match(test_two_level_dict, "foo:bar:baz")
    # In test_two_level_comb_dict, 'foo:bar' corresponds to 'baz:woz', not
    # 'baz'. This match should return False.
    assert not salt.utils.data.subdict_match(test_two_level_comb_dict, "foo:bar:baz")
    # This tests matching with the delimiter in the value part (in other
    # words, that the path 'foo:bar' corresponds to the string 'baz:woz').
    assert salt.utils.data.subdict_match(test_two_level_comb_dict, "foo:bar:baz:woz")
    # This would match if test_two_level_comb_dict['foo']['bar'] was equal
    # to 'baz:woz:wiz', or if there was more deep nesting. But it does not,
    # so this should return False.
    assert not salt.utils.data.subdict_match(
        test_two_level_comb_dict, "foo:bar:baz:woz:wiz"
    )
    # This tests for cases when a key path corresponds to a list. The
    # value part 'ghi' should be successfully matched as it is a member of
    # the list corresponding to key path 'abc'. It is somewhat a
    # duplication of a test within test_traverse_dict_and_list, but
    # salt.utils.data.subdict_match() does more than just invoke
    # salt.utils.traverse_list_and_dict() so this particular assertion is a
    # sanity check.
    assert salt.utils.data.subdict_match(test_two_level_dict_and_list, "abc:ghi")
    # This tests the use case of a dict embedded in a list, embedded in a
    # list, embedded in a dict. This is a rather absurd case, but it
    # confirms that match recursion works properly.
    assert salt.utils.data.subdict_match(
        test_two_level_dict_and_list, "abc:lorem:ipsum:dolor:sit"
    )
    # Test four level dict match for reference
    assert salt.utils.data.subdict_match(test_three_level_dict, "a:b:c:v")
    # Test regression in 2015.8 where 'a:c:v' would match 'a:b:c:v'
    assert not salt.utils.data.subdict_match(test_three_level_dict, "a:c:v")
    # Test wildcard match
    assert salt.utils.data.subdict_match(test_three_level_dict, "a:*:c:v")


@pytest.mark.parametrize(
    "wildcard",
    [
        ("*:*:*:*"),
        ("a:*:*:*"),
        ("a:b:*:*"),
        ("a:b:ç:*"),
        ("a:b:*:d"),
        ("a:*:ç:d"),
        ("*:b:ç:d"),
        ("*:*:ç:d"),
        ("*:*:*:d"),
        ("a:*:*:d"),
        ("a:b:*:ef*"),
        ("a:b:*:g*"),
        ("a:b:*:j:*"),
        ("a:b:*:j:k"),
        ("a:b:*:*:k"),
        ("a:b:*:*:*"),
    ],
)
def test_subdict_match_with_wildcards(wildcard):
    """
    Tests subdict matching when wildcards are used in the expression
    """
    data = {"a": {"b": {"ç": "d", "é": ["eff", "gee", "8ch"], "ĩ": {"j": "k"}}}}
    assert salt.utils.data.subdict_match(data, wildcard)


def test_traverse_dict():
    test_two_level_dict = {"foo": {"bar": "baz"}}

    assert {"not_found": "nope"} == salt.utils.data.traverse_dict(
        test_two_level_dict, "foo:bar:baz", {"not_found": "nope"}
    )
    assert "baz" == salt.utils.data.traverse_dict(
        test_two_level_dict, "foo:bar", {"not_found": "not_found"}
    )


def test_traverse_dict_and_list():
    test_two_level_dict = {"foo": {"bar": "baz"}}
    test_two_level_dict_and_list = {
        "foo": ["bar", "baz", {"lorem": {"ipsum": [{"dolor": "sit"}]}}]
    }

    # Check traversing too far: salt.utils.data.traverse_dict_and_list() returns
    # the value corresponding to a given key path, and baz is a value
    # corresponding to the key path foo:bar.
    assert {"not_found": "nope"} == salt.utils.data.traverse_dict_and_list(
        test_two_level_dict, "foo:bar:baz", {"not_found": "nope"}
    )
    # Now check to ensure that foo:bar corresponds to baz
    assert "baz" == salt.utils.data.traverse_dict_and_list(
        test_two_level_dict, "foo:bar", {"not_found": "not_found"}
    )
    # Check traversing too far
    assert {"not_found": "nope"} == salt.utils.data.traverse_dict_and_list(
        test_two_level_dict_and_list, "foo:bar", {"not_found": "nope"}
    )
    # Check index 1 (2nd element) of list corresponding to path 'foo'
    assert "baz" == salt.utils.data.traverse_dict_and_list(
        test_two_level_dict_and_list, "foo:1", {"not_found": "not_found"}
    )
    # Traverse a couple times into dicts embedded in lists
    assert "sit" == salt.utils.data.traverse_dict_and_list(
        test_two_level_dict_and_list,
        "foo:lorem:ipsum:dolor",
        {"not_found": "not_found"},
    )

    # Traverse and match integer key in a nested dict
    # https://github.com/saltstack/salt/issues/56444
    assert "it worked" == salt.utils.data.traverse_dict_and_list(
        {"foo": {1234: "it worked"}},
        "foo:1234",
        "it didn't work",
    )
    # Make sure that we properly return the default value when the initial
    # attempt fails and YAML-loading the target key doesn't change its
    # value.
    assert "default" == salt.utils.data.traverse_dict_and_list(
        {"foo": {"baz": "didn't work"}},
        "foo:bar",
        "default",
    )


def test_issue_39709():
    test_two_level_dict_and_list = {
        "foo": ["bar", "baz", {"lorem": {"ipsum": [{"dolor": "sit"}]}}]
    }

    assert "sit" == salt.utils.data.traverse_dict_and_list(
        test_two_level_dict_and_list,
        ["foo", "lorem", "ipsum", "dolor"],
        {"not_found": "not_found"},
    )


def test_compare_dicts():
    ret = salt.utils.data.compare_dicts(old={"foo": "bar"}, new={"foo": "bar"})
    assert ret == {}

    ret = salt.utils.data.compare_dicts(old={"foo": "bar"}, new={"foo": "woz"})
    expected_ret = {"foo": {"new": "woz", "old": "bar"}}
    assert ret == expected_ret


def test_compare_lists_no_change():
    ret = salt.utils.data.compare_lists(
        old=[1, 2, 3, "a", "b", "c"], new=[1, 2, 3, "a", "b", "c"]
    )
    expected = {}
    assert ret == expected


def test_compare_lists_changes():
    ret = salt.utils.data.compare_lists(
        old=[1, 2, 3, "a", "b", "c"], new=[1, 2, 4, "x", "y", "z"]
    )
    expected = {"new": [4, "x", "y", "z"], "old": [3, "a", "b", "c"]}
    assert ret == expected


def test_compare_lists_changes_new():
    ret = salt.utils.data.compare_lists(old=[1, 2, 3], new=[1, 2, 3, "x", "y", "z"])
    expected = {"new": ["x", "y", "z"]}
    assert ret == expected


def test_compare_lists_changes_old():
    ret = salt.utils.data.compare_lists(old=[1, 2, 3, "a", "b", "c"], new=[1, 2, 3])
    expected = {"old": ["a", "b", "c"]}
    assert ret == expected


def test_decode(get_test_data, get_BYTES, get_EGGS):
    """
    Companion to test_decode_to_str, they should both be kept up-to-date
    with one another.

    NOTE: This uses the lambda "_b" defined above in the global scope,
    which encodes a string to a bytestring, assuming utf-8.
    """
    expected = [
        "unicode_str",
        "питон",
        123,
        456.789,
        True,
        False,
        None,
        "яйца",
        get_BYTES,
        [123, 456.789, "спам", True, False, None, "яйца", get_BYTES],
        (987, 654.321, "яйца", "яйца", None, (True, "яйца", get_BYTES)),
        {
            "str_key": "str_val",
            None: True,
            123: 456.789,
            "яйца": get_BYTES,
            "subdict": {
                "unicode_key": "яйца",
                "tuple": (123, "hello", "world", True, "яйца", get_BYTES),
                "list": [456, "спам", False, "яйца", get_BYTES],
            },
        },
        SaltOrderedDict([("foo", "bar"), (123, 456), ("яйца", get_BYTES)]),
    ]

    ret = salt.utils.data.decode(
        get_test_data,
        keep=True,
        normalize=True,
        preserve_dict_class=True,
        preserve_tuples=True,
    )
    assert ret == expected

    # The binary data in the data structure should fail to decode, even
    # using the fallback, and raise an exception.
    pytest.raises(
        UnicodeDecodeError,
        salt.utils.data.decode,
        get_test_data,
        keep=False,
        normalize=True,
        preserve_dict_class=True,
        preserve_tuples=True,
    )

    # Now munge the expected data so that we get what we would expect if we
    # disable preservation of dict class and tuples
    expected[10] = [987, 654.321, "яйца", "яйца", None, [True, "яйца", get_BYTES]]
    expected[11]["subdict"]["tuple"] = [123, "hello", "world", True, "яйца", get_BYTES]
    expected[12] = {"foo": "bar", 123: 456, "яйца": get_BYTES}

    ret = salt.utils.data.decode(
        get_test_data,
        keep=True,
        normalize=True,
        preserve_dict_class=False,
        preserve_tuples=False,
    )
    assert ret == expected

    # Now test single non-string, non-data-structure items, these should
    # return the same value when passed to this function
    for item in (123, 4.56, True, False, None):
        log.debug("Testing decode of %s", item)
        assert salt.utils.data.decode(item) == item

    # Test single strings (not in a data structure)
    assert salt.utils.data.decode("foo") == "foo"
    assert salt.utils.data.decode(_b("bar")) == "bar"
    assert salt.utils.data.decode(get_EGGS, normalize=True) == "яйца"
    assert salt.utils.data.decode(get_EGGS, normalize=False) == get_EGGS

    # Test binary blob
    assert salt.utils.data.decode(get_BYTES, keep=True) == get_BYTES
    pytest.raises(UnicodeDecodeError, salt.utils.data.decode, get_BYTES, keep=False)


def test_circular_refs_dicts():
    test_dict = {"key": "value", "type": "test1"}
    test_dict["self"] = test_dict
    ret = salt.utils.data._remove_circular_refs(ob=test_dict)
    assert ret == {"key": "value", "type": "test1", "self": None}


def test_circular_refs_lists():
    test_list = {
        "foo": [],
    }
    test_list["foo"].append((test_list,))
    ret = salt.utils.data._remove_circular_refs(ob=test_list)
    assert ret == {"foo": [(None,)]}


def test_circular_refs_tuple():
    test_dup = {"foo": "string 1", "bar": "string 1", "ham": 1, "spam": 1}
    ret = salt.utils.data._remove_circular_refs(ob=test_dup)
    assert ret == {"foo": "string 1", "bar": "string 1", "ham": 1, "spam": 1}


def test_decode_to_str(get_test_data, get_BYTES):
    """
    Companion to test_decode, they should both be kept up-to-date with one
    another.

    NOTE: This uses the lambda "_s" defined above in the global scope,
    which converts the string/bytestring to a str type.
    """
    expected = [
        _s("unicode_str"),
        _s("питон"),
        123,
        456.789,
        True,
        False,
        None,
        _s("яйца"),
        get_BYTES,
        [123, 456.789, _s("спам"), True, False, None, _s("яйца"), get_BYTES],
        (987, 654.321, _s("яйца"), _s("яйца"), None, (True, _s("яйца"), get_BYTES)),
        {
            _s("str_key"): _s("str_val"),
            None: True,
            123: 456.789,
            _s("яйца"): get_BYTES,
            _s("subdict"): {
                _s("unicode_key"): _s("яйца"),
                _s("tuple"): (
                    123,
                    _s("hello"),
                    _s("world"),
                    True,
                    _s("яйца"),
                    get_BYTES,
                ),
                _s("list"): [456, _s("спам"), False, _s("яйца"), get_BYTES],
            },
        },
        SaltOrderedDict([(_s("foo"), _s("bar")), (123, 456), (_s("яйца"), get_BYTES)]),
    ]

    ret = salt.utils.data.decode(
        get_test_data,
        keep=True,
        normalize=True,
        preserve_dict_class=True,
        preserve_tuples=True,
        to_str=True,
    )
    assert ret == expected

    # The binary data in the data structure should fail to decode, even
    # using the fallback, and raise an exception.
    pytest.raises(
        UnicodeDecodeError,
        salt.utils.data.decode,
        get_test_data,
        keep=False,
        normalize=True,
        preserve_dict_class=True,
        preserve_tuples=True,
        to_str=True,
    )

    # Now munge the expected data so that we get what we would expect if we
    # disable preservation of dict class and tuples
    expected[10] = [
        987,
        654.321,
        _s("яйца"),
        _s("яйца"),
        None,
        [True, _s("яйца"), get_BYTES],
    ]
    expected[11][_s("subdict")][_s("tuple")] = [
        123,
        _s("hello"),
        _s("world"),
        True,
        _s("яйца"),
        get_BYTES,
    ]
    expected[12] = {_s("foo"): _s("bar"), 123: 456, _s("яйца"): get_BYTES}

    ret = salt.utils.data.decode(
        get_test_data,
        keep=True,
        normalize=True,
        preserve_dict_class=False,
        preserve_tuples=False,
        to_str=True,
    )
    assert ret == expected

    # Now test single non-string, non-data-structure items, these should
    # return the same value when passed to this function
    for item in (123, 4.56, True, False, None):
        log.debug("Testing decode of %s", item)
        assert salt.utils.data.decode(item, to_str=True) == item

    # Test single strings (not in a data structure)
    assert salt.utils.data.decode("foo", to_str=True) == _s("foo")
    assert salt.utils.data.decode(_b("bar"), to_str=True) == _s("bar")

    # Test binary blob
    assert salt.utils.data.decode(get_BYTES, keep=True, to_str=True) == get_BYTES
    pytest.raises(
        UnicodeDecodeError,
        salt.utils.data.decode,
        get_BYTES,
        keep=False,
        to_str=True,
    )


def test_decode_fallback():
    """
    Test fallback to utf-8
    """
    with patch.object(builtins, "__salt_system_encoding__", "ascii"):
        assert salt.utils.data.decode(_b("яйца")) == "яйца"


def test_encode(get_test_data, get_BYTES, get_EGGS):
    """
    NOTE: This uses the lambda "_b" defined above in the global scope,
    which encodes a string to a bytestring, assuming utf-8.
    """
    expected = [
        _b("unicode_str"),
        _b("питон"),
        123,
        456.789,
        True,
        False,
        None,
        _b(get_EGGS),
        get_BYTES,
        [123, 456.789, _b("спам"), True, False, None, _b(get_EGGS), get_BYTES],
        (987, 654.321, _b("яйца"), _b(get_EGGS), None, (True, _b(get_EGGS), get_BYTES)),
        {
            _b("str_key"): _b("str_val"),
            None: True,
            123: 456.789,
            _b(get_EGGS): get_BYTES,
            _b("subdict"): {
                _b("unicode_key"): _b(get_EGGS),
                _b("tuple"): (
                    123,
                    _b("hello"),
                    _b("world"),
                    True,
                    _b(get_EGGS),
                    get_BYTES,
                ),
                _b("list"): [456, _b("спам"), False, _b(get_EGGS), get_BYTES],
            },
        },
        SaltOrderedDict(
            [(_b("foo"), _b("bar")), (123, 456), (_b(get_EGGS), get_BYTES)]
        ),
    ]

    # Both keep=True and keep=False should work because the get_BYTES data is
    # already bytes.
    ret = salt.utils.data.encode(
        get_test_data, keep=True, preserve_dict_class=True, preserve_tuples=True
    )
    assert ret == expected
    ret = salt.utils.data.encode(
        get_test_data, keep=False, preserve_dict_class=True, preserve_tuples=True
    )
    assert ret == expected

    # Now munge the expected data so that we get what we would expect if we
    # disable preservation of dict class and tuples
    expected[10] = [
        987,
        654.321,
        _b("яйца"),
        _b(get_EGGS),
        None,
        [True, _b(get_EGGS), get_BYTES],
    ]
    expected[11][_b("subdict")][  # pylint: disable=unsupported-assignment-operation
        _b("tuple")
    ] = [
        123,
        _b("hello"),
        _b("world"),
        True,
        _b(get_EGGS),
        get_BYTES,
    ]
    expected[12] = {_b("foo"): _b("bar"), 123: 456, _b(get_EGGS): get_BYTES}

    ret = salt.utils.data.encode(
        get_test_data, keep=True, preserve_dict_class=False, preserve_tuples=False
    )
    assert ret == expected
    ret = salt.utils.data.encode(
        get_test_data, keep=False, preserve_dict_class=False, preserve_tuples=False
    )
    assert ret == expected

    # Now test single non-string, non-data-structure items, these should
    # return the same value when passed to this function
    for item in (123, 4.56, True, False, None):
        log.debug("Testing encode of %s", item)
        assert salt.utils.data.encode(item) == item

    # Test single strings (not in a data structure)
    assert salt.utils.data.encode("foo") == _b("foo")
    assert salt.utils.data.encode(_b("bar")) == _b("bar")

    # Test binary blob, nothing should happen even when keep=False since
    # the data is already bytes
    assert salt.utils.data.encode(get_BYTES, keep=True) == get_BYTES
    assert salt.utils.data.encode(get_BYTES, keep=False) == get_BYTES


def test_encode_keep():
    """
    Whereas we tested the keep argument in test_decode, it is much easier
    to do a more comprehensive test of keep in its own function where we
    can force the encoding.
    """
    unicode_str = "питон"
    encoding = "ascii"

    # Test single string
    assert salt.utils.data.encode(unicode_str, encoding, keep=True) == unicode_str
    pytest.raises(
        UnicodeEncodeError,
        salt.utils.data.encode,
        unicode_str,
        encoding,
        keep=False,
    )

    data = [
        unicode_str,
        [b"foo", [unicode_str], {b"key": unicode_str}, (unicode_str,)],
        {
            b"list": [b"foo", unicode_str],
            b"dict": {b"key": unicode_str},
            b"tuple": (b"foo", unicode_str),
        },
        ([b"foo", unicode_str], {b"key": unicode_str}, (unicode_str,)),
    ]

    # Since everything was a bytestring aside from the bogus data, the
    # return data should be identical. We don't need to test recursive
    # decoding, that has already been tested in test_encode.
    assert (
        salt.utils.data.encode(data, encoding, keep=True, preserve_tuples=True) == data
    )
    pytest.raises(
        UnicodeEncodeError,
        salt.utils.data.encode,
        data,
        encoding,
        keep=False,
        preserve_tuples=True,
    )

    for index, _ in enumerate(data):
        assert (
            salt.utils.data.encode(
                data[index], encoding, keep=True, preserve_tuples=True
            )
            == data[index]
        )
        pytest.raises(
            UnicodeEncodeError,
            salt.utils.data.encode,
            data[index],
            encoding,
            keep=False,
            preserve_tuples=True,
        )


def test_encode_fallback():
    """
    Test fallback to utf-8
    """
    with patch.object(builtins, "__salt_system_encoding__", "ascii"):
        assert salt.utils.data.encode("яйца") == _b("яйца")
    with patch.object(builtins, "__salt_system_encoding__", "CP1252"):
        assert salt.utils.data.encode("Ψ") == _b("Ψ")


def test_repack_dict():
    list_of_one_element_dicts = [
        {"dict_key_1": "dict_val_1"},
        {"dict_key_2": "dict_val_2"},
        {"dict_key_3": "dict_val_3"},
    ]
    expected_ret = {
        "dict_key_1": "dict_val_1",
        "dict_key_2": "dict_val_2",
        "dict_key_3": "dict_val_3",
    }
    ret = salt.utils.data.repack_dictlist(list_of_one_element_dicts)
    assert ret == expected_ret

    # Try with yaml
    yaml_key_val_pair = "- key1: val1"
    ret = salt.utils.data.repack_dictlist(yaml_key_val_pair)
    assert ret == {"key1": "val1"}

    # Make sure we handle non-yaml junk data
    ret = salt.utils.data.repack_dictlist(LOREM_IPSUM)
    assert ret == {}


def test_stringify():
    pytest.raises(TypeError, salt.utils.data.stringify, 9)
    assert salt.utils.data.stringify(["one", "two", "three", 4, 5]) == [
        "one",
        "two",
        "three",
        "4",
        "5",
    ]


def test_json_query():
    # Raises exception if jmespath module is not found
    with patch("salt.utils.data.jmespath", None):
        with pytest.raises(RuntimeError, match="requires jmespath"):
            salt.utils.data.json_query({}, "@")

    # Test search
    user_groups = {
        "user1": {"groups": ["group1", "group2", "group3"]},
        "user2": {"groups": ["group1", "group2"]},
        "user3": {"groups": ["group3"]},
    }
    expression = "*.groups[0]"
    primary_groups = ["group1", "group1", "group3"]
    assert sorted(salt.utils.data.json_query(user_groups, expression)) == primary_groups


def test_nop():
    """
    Test cases where nothing will be done.
    """
    # Test with dictionary without recursion
    old_dict = {
        "foo": "bar",
        "bar": {"baz": {"qux": "quux"}},
        "baz": ["qux", {"foo": "bar"}],
    }
    new_dict = salt.utils.data.filter_falsey(old_dict)
    assert old_dict == new_dict
    # Check returned type equality
    assert type(old_dict) is type(new_dict)
    # Test dictionary with recursion
    new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=3)
    assert old_dict == new_dict
    # Test with list
    old_list = ["foo", "bar"]
    new_list = salt.utils.data.filter_falsey(old_list)
    assert old_list == new_list
    # Check returned type equality
    assert type(old_list) is type(new_list)
    # Test with set
    old_set = {"foo", "bar"}
    new_set = salt.utils.data.filter_falsey(old_set)
    assert old_set == new_set
    # Check returned type equality
    assert type(old_set) is type(new_set)
    # Test with SaltOrderedDict
    old_dict = SaltOrderedDict(
        [
            ("foo", "bar"),
            ("bar", SaltOrderedDict([("qux", "quux")])),
            ("baz", ["qux", SaltOrderedDict([("foo", "bar")])]),
        ]
    )
    new_dict = salt.utils.data.filter_falsey(old_dict)
    assert old_dict == new_dict
    assert type(old_dict) is type(new_dict)
    # Test excluding int
    old_list = [0]
    new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[int])
    assert old_list == new_list
    # Test excluding str (or unicode) (or both)
    old_list = [""]
    new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[str])
    assert old_list == new_list
    # Test excluding list
    old_list = [[]]
    new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[type([])])
    assert old_list == new_list
    # Test excluding dict
    old_list = [{}]
    new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[type({})])
    assert old_list == new_list


def test_filter_dict_no_recurse():
    """
    Test filtering a dictionary without recursing.
    This will only filter out key-values where the values are falsey.
    """
    old_dict = {
        "foo": None,
        "bar": {"baz": {"qux": None, "quux": "", "foo": []}},
        "baz": ["qux"],
        "qux": {},
        "quux": [],
    }
    new_dict = salt.utils.data.filter_falsey(old_dict)
    expect_dict = {
        "bar": {"baz": {"qux": None, "quux": "", "foo": []}},
        "baz": ["qux"],
    }
    assert expect_dict == new_dict
    assert type(expect_dict) is type(new_dict)


def test_filter_dict_recurse():
    """
    Test filtering a dictionary with recursing.
    This will filter out any key-values where the values are falsey or when
    the values *become* falsey after filtering their contents (in case they
    are lists or dicts).
    """
    old_dict = {
        "foo": None,
        "bar": {"baz": {"qux": None, "quux": "", "foo": []}},
        "baz": ["qux"],
        "qux": {},
        "quux": [],
    }
    new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=3)
    expect_dict = {"baz": ["qux"]}
    assert expect_dict == new_dict
    assert type(expect_dict) is type(new_dict)


def test_filter_list_no_recurse():
    """
    Test filtering a list without recursing.
    This will only filter out items which are falsey.
    """
    old_list = ["foo", None, [], {}, 0, ""]
    new_list = salt.utils.data.filter_falsey(old_list)
    expect_list = ["foo"]
    assert expect_list == new_list
    assert type(expect_list) is type(new_list)
    # Ensure nested values are *not* filtered out.
    old_list = [
        "foo",
        ["foo"],
        ["foo", None],
        {"foo": 0},
        {"foo": "bar", "baz": []},
        [{"foo": ""}],
    ]
    new_list = salt.utils.data.filter_falsey(old_list)
    assert old_list == new_list
    assert type(old_list) is type(new_list)


def test_filter_list_recurse():
    """
    Test filtering a list with recursing.
    This will filter out any items which are falsey, or which become falsey
    after filtering their contents (in case they are lists or dicts).
    """
    old_list = [
        "foo",
        ["foo"],
        ["foo", None],
        {"foo": 0},
        {"foo": "bar", "baz": []},
        [{"foo": ""}],
    ]
    new_list = salt.utils.data.filter_falsey(old_list, recurse_depth=3)
    expect_list = ["foo", ["foo"], ["foo"], {"foo": "bar"}]
    assert expect_list == new_list
    assert type(expect_list) is type(new_list)


def test_filter_set_no_recurse():
    """
    Test filtering a set without recursing.
    Note that a set cannot contain unhashable types, so recursion is not possible.
    """
    old_set = {"foo", None, 0, ""}
    new_set = salt.utils.data.filter_falsey(old_set)
    expect_set = {"foo"}
    assert expect_set == new_set
    assert type(expect_set) is type(new_set)


def test_filter_ordereddict_no_recurse():
    """
    Test filtering an SaltOrderedDict without recursing.
    """
    old_dict = SaltOrderedDict(
        [
            ("foo", None),
            (
                "bar",
                SaltOrderedDict(
                    [
                        (
                            "baz",
                            SaltOrderedDict([("qux", None), ("quux", ""), ("foo", [])]),
                        )
                    ]
                ),
            ),
            ("baz", ["qux"]),
            ("qux", {}),
            ("quux", []),
        ]
    )
    new_dict = salt.utils.data.filter_falsey(old_dict)
    expect_dict = SaltOrderedDict(
        [
            (
                "bar",
                SaltOrderedDict(
                    [
                        (
                            "baz",
                            SaltOrderedDict([("qux", None), ("quux", ""), ("foo", [])]),
                        )
                    ]
                ),
            ),
            ("baz", ["qux"]),
        ]
    )
    assert expect_dict == new_dict
    assert type(expect_dict) is type(new_dict)


def test_filter_ordereddict_recurse():
    """
    Test filtering an SaltOrderedDict with recursing.
    """
    old_dict = SaltOrderedDict(
        [
            ("foo", None),
            (
                "bar",
                SaltOrderedDict(
                    [
                        (
                            "baz",
                            SaltOrderedDict([("qux", None), ("quux", ""), ("foo", [])]),
                        )
                    ]
                ),
            ),
            ("baz", ["qux"]),
            ("qux", {}),
            ("quux", []),
        ]
    )
    new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=3)
    expect_dict = SaltOrderedDict([("baz", ["qux"])])
    assert expect_dict == new_dict
    assert type(expect_dict) is type(new_dict)


def test_filter_list_recurse_limit():
    """
    Test filtering a list with recursing, but with a limited depth.
    Note that the top-level is always processed, so a recursion depth of 2
    means that two *additional* levels are processed.
    """
    old_list = [None, [None, [None, [None]]]]
    new_list = salt.utils.data.filter_falsey(old_list, recurse_depth=2)
    assert [[[[None]]]] == new_list


def test_filter_dict_recurse_limit():
    """
    Test filtering a dict with recursing, but with a limited depth.
    Note that the top-level is always processed, so a recursion depth of 2
    means that two *additional* levels are processed.
    """
    old_dict = {
        "one": None,
        "foo": {"two": None, "bar": {"three": None, "baz": {"four": None}}},
    }
    new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=2)
    assert {"foo": {"bar": {"baz": {"four": None}}}} == new_dict


def test_filter_exclude_types():
    """
    Test filtering a list recursively, but also ignoring (i.e. not filtering)
    out certain types that can be falsey.
    """
    # Ignore int, unicode
    old_list = [
        "foo",
        ["foo"],
        ["foo", None],
        {"foo": 0},
        {"foo": "bar", "baz": []},
        [{"foo": ""}],
    ]
    new_list = salt.utils.data.filter_falsey(
        old_list, recurse_depth=3, ignore_types=[int, str]
    )
    assert [
        "foo",
        ["foo"],
        ["foo"],
        {"foo": 0},
        {"foo": "bar"},
        [{"foo": ""}],
    ] == new_list
    # Ignore list
    old_list = [
        "foo",
        ["foo"],
        ["foo", None],
        {"foo": 0},
        {"foo": "bar", "baz": []},
        [{"foo": ""}],
    ]
    new_list = salt.utils.data.filter_falsey(
        old_list, recurse_depth=3, ignore_types=[type([])]
    )
    assert ["foo", ["foo"], ["foo"], {"foo": "bar", "baz": []}, []] == new_list
    # Ignore dict
    old_list = [
        "foo",
        ["foo"],
        ["foo", None],
        {"foo": 0},
        {"foo": "bar", "baz": []},
        [{"foo": ""}],
    ]
    new_list = salt.utils.data.filter_falsey(
        old_list, recurse_depth=3, ignore_types=[type({})]
    )
    assert ["foo", ["foo"], ["foo"], {}, {"foo": "bar"}, [{}]] == new_list
    # Ignore NoneType
    old_list = [
        "foo",
        ["foo"],
        ["foo", None],
        {"foo": 0},
        {"foo": "bar", "baz": []},
        [{"foo": ""}],
    ]
    new_list = salt.utils.data.filter_falsey(
        old_list, recurse_depth=3, ignore_types=[type(None)]
    )
    assert ["foo", ["foo"], ["foo", None], {"foo": "bar"}] == new_list


def test_list_equality():
    """
    Test cases where equal lists are compared.
    """
    test_list = [0, 1, 2]
    assert {} == salt.utils.data.recursive_diff(test_list, test_list)

    test_list = [[0], [1], [0, 1, 2]]
    assert {} == salt.utils.data.recursive_diff(test_list, test_list)


def test_dict_equality():
    """
    Test cases where equal dicts are compared.
    """
    test_dict = {"foo": "bar", "bar": {"baz": {"qux": "quux"}}, "frop": 0}
    assert {} == salt.utils.data.recursive_diff(test_dict, test_dict)


def test_ordereddict_equality():
    """
    Test cases where equal SaltOrderedDicts are compared.
    """
    test_dict = SaltOrderedDict(
        [
            ("foo", "bar"),
            ("bar", SaltOrderedDict([("baz", SaltOrderedDict([("qux", "quux")]))])),
            ("frop", 0),
        ]
    )
    assert {} == salt.utils.data.recursive_diff(test_dict, test_dict)


def test_mixed_equality():
    """
    Test cases where mixed nested lists and dicts are compared.
    """
    test_data = {
        "foo": "bar",
        "baz": [0, 1, 2],
        "bar": {"baz": [{"qux": "quux"}, {"froop", 0}]},
    }
    assert {} == salt.utils.data.recursive_diff(test_data, test_data)


def test_set_equality():
    """
    Test cases where equal sets are compared.
    """
    test_set = {0, 1, 2, 3, "foo"}
    assert {} == salt.utils.data.recursive_diff(test_set, test_set)

    # This is a bit of an oddity, as python seems to sort the sets in memory
    # so both sets end up with the same ordering (0..3).
    set_one = {0, 1, 2, 3}
    set_two = {3, 2, 1, 0}
    assert {} == salt.utils.data.recursive_diff(set_one, set_two)


def test_tuple_equality():
    """
    Test cases where equal tuples are compared.
    """
    test_tuple = (0, 1, 2, 3, "foo")
    assert {} == salt.utils.data.recursive_diff(test_tuple, test_tuple)


def test_list_inequality():
    """
    Test cases where two inequal lists are compared.
    """
    list_one = [0, 1, 2]
    list_two = ["foo", "bar", "baz"]
    expected_result = {"old": list_one, "new": list_two}
    assert expected_result == salt.utils.data.recursive_diff(list_one, list_two)
    expected_result = {"new": list_one, "old": list_two}
    assert expected_result == salt.utils.data.recursive_diff(list_two, list_one)

    list_one = [0, "foo", 1, "bar"]
    list_two = [1, "foo", 1, "qux"]
    expected_result = {"old": [0, "bar"], "new": [1, "qux"]}
    assert expected_result == salt.utils.data.recursive_diff(list_one, list_two)
    expected_result = {"new": [0, "bar"], "old": [1, "qux"]}
    assert expected_result == salt.utils.data.recursive_diff(list_two, list_one)

    list_one = [0, 1, [2, 3]]
    list_two = [0, 1, ["foo", "bar"]]
    expected_result = {"old": [[2, 3]], "new": [["foo", "bar"]]}
    assert expected_result == salt.utils.data.recursive_diff(list_one, list_two)
    expected_result = {"new": [[2, 3]], "old": [["foo", "bar"]]}
    assert expected_result == salt.utils.data.recursive_diff(list_two, list_one)


def test_dict_inequality():
    """
    Test cases where two inequal dicts are compared.
    """
    dict_one = {"foo": 1, "bar": 2, "baz": 3}
    dict_two = {"foo": 2, 1: "bar", "baz": 3}
    expected_result = {"old": {"foo": 1, "bar": 2}, "new": {"foo": 2, 1: "bar"}}
    assert expected_result == salt.utils.data.recursive_diff(dict_one, dict_two)
    expected_result = {"new": {"foo": 1, "bar": 2}, "old": {"foo": 2, 1: "bar"}}
    assert expected_result == salt.utils.data.recursive_diff(dict_two, dict_one)

    dict_one = {"foo": {"bar": {"baz": 1}}}
    dict_two = {"foo": {"qux": {"baz": 1}}}
    expected_result = {"old": dict_one, "new": dict_two}
    assert expected_result == salt.utils.data.recursive_diff(dict_one, dict_two)
    expected_result = {"new": dict_one, "old": dict_two}
    assert expected_result == salt.utils.data.recursive_diff(dict_two, dict_one)


def test_ordereddict_inequality():
    """
    Test cases where two inequal SaltOrderedDicts are compared.
    """
    odict_one = SaltOrderedDict([("foo", "bar"), ("bar", "baz")])
    odict_two = SaltOrderedDict([("bar", "baz"), ("foo", "bar")])
    expected_result = {"old": odict_one, "new": odict_two}
    assert expected_result == salt.utils.data.recursive_diff(odict_one, odict_two)


def test_set_inequality():
    """
    Test cases where two inequal sets are compared.
    Tricky as the sets are compared zipped, so shuffled sets of equal values
    are considered different.
    """
    set_one = {0, 1, 2, 4}
    set_two = {0, 1, 3, 4}
    expected_result = {"old": {2}, "new": {3}}
    assert expected_result == salt.utils.data.recursive_diff(set_one, set_two)
    expected_result = {"new": {2}, "old": {3}}
    assert expected_result == salt.utils.data.recursive_diff(set_two, set_one)

    # It is unknown how different python versions will store sets in memory.
    # Python 2.7 seems to sort it (i.e. set_one below becomes {0, 1, 'foo', 'bar'}
    # However Python 3.6.8 stores it differently each run.
    # So just test for "not equal" here.
    set_one = {0, "foo", 1, "bar"}
    set_two = {"foo", 1, "bar", 2}
    expected_result = {}
    assert expected_result != salt.utils.data.recursive_diff(set_one, set_two)


def test_mixed_inequality():
    """
    Test cases where two mixed dicts/iterables that are different are compared.
    """
    dict_one = {"foo": [1, 2, 3]}
    dict_two = {"foo": [3, 2, 1]}
    expected_result = {"old": {"foo": [1, 3]}, "new": {"foo": [3, 1]}}
    assert expected_result == salt.utils.data.recursive_diff(dict_one, dict_two)
    expected_result = {"new": {"foo": [1, 3]}, "old": {"foo": [3, 1]}}
    assert expected_result == salt.utils.data.recursive_diff(dict_two, dict_one)

    list_one = [1, 2, {"foo": ["bar", {"foo": 1, "bar": 2}]}]
    list_two = [3, 4, {"foo": ["qux", {"foo": 1, "bar": 2}]}]
    expected_result = {
        "old": [1, 2, {"foo": ["bar"]}],
        "new": [3, 4, {"foo": ["qux"]}],
    }
    assert expected_result == salt.utils.data.recursive_diff(list_one, list_two)
    expected_result = {
        "new": [1, 2, {"foo": ["bar"]}],
        "old": [3, 4, {"foo": ["qux"]}],
    }
    assert expected_result == salt.utils.data.recursive_diff(list_two, list_one)

    mixed_one = {"foo": {0, 1, 2}, "bar": [0, 1, 2]}
    mixed_two = {"foo": {1, 2, 3}, "bar": [1, 2, 3]}
    expected_result = {
        "old": {"foo": {0}, "bar": [0, 1, 2]},
        "new": {"foo": {3}, "bar": [1, 2, 3]},
    }
    assert expected_result == salt.utils.data.recursive_diff(mixed_one, mixed_two)
    expected_result = {
        "new": {"foo": {0}, "bar": [0, 1, 2]},
        "old": {"foo": {3}, "bar": [1, 2, 3]},
    }
    assert expected_result == salt.utils.data.recursive_diff(mixed_two, mixed_one)


def test_tuple_inequality():
    """
    Test cases where two tuples that are different are compared.
    """
    tuple_one = (1, 2, 3)
    tuple_two = (3, 2, 1)
    expected_result = {"old": (1, 3), "new": (3, 1)}
    assert expected_result == salt.utils.data.recursive_diff(tuple_one, tuple_two)


def test_list_vs_set():
    """
    Test case comparing a list with a set, will be compared unordered.
    """
    mixed_one = [1, 2, 3]
    mixed_two = {3, 2, 1}
    expected_result = {}
    assert expected_result == salt.utils.data.recursive_diff(mixed_one, mixed_two)
    assert expected_result == salt.utils.data.recursive_diff(mixed_two, mixed_one)


def test_dict_vs_ordereddict():
    """
    Test case comparing a dict with an ordereddict, will be compared unordered.
    """
    test_dict = {"foo": "bar", "bar": "baz"}
    test_odict = SaltOrderedDict([("foo", "bar"), ("bar", "baz")])
    assert {} == salt.utils.data.recursive_diff(test_dict, test_odict)
    assert {} == salt.utils.data.recursive_diff(test_odict, test_dict)

    test_odict2 = SaltOrderedDict([("bar", "baz"), ("foo", "bar")])
    assert {} == salt.utils.data.recursive_diff(test_dict, test_odict2)
    assert {} == salt.utils.data.recursive_diff(test_odict2, test_dict)


def test_list_ignore_ignored():
    """
    Test case comparing two lists with ignore-list supplied (which is not used
    when comparing lists).
    """
    list_one = [1, 2, 3]
    list_two = [3, 2, 1]
    expected_result = {"old": [1, 3], "new": [3, 1]}
    assert expected_result == salt.utils.data.recursive_diff(
        list_one, list_two, ignore_keys=[1, 3]
    )


def test_dict_ignore():
    """
    Test case comparing two dicts with ignore-list supplied.
    """
    dict_one = {"foo": 1, "bar": 2, "baz": 3}
    dict_two = {"foo": 3, "bar": 2, "baz": 1}
    expected_result = {"old": {"baz": 3}, "new": {"baz": 1}}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_keys=["foo"]
    )


def test_ordereddict_ignore():
    """
    Test case comparing two SaltOrderedDicts with ignore-list supplied.
    """
    odict_one = SaltOrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
    odict_two = SaltOrderedDict([("baz", 1), ("bar", 2), ("foo", 3)])
    # The key 'foo' will be ignored, which means the key from the other SaltOrderedDict
    # will always be considered "different" since SaltOrderedDicts are compared ordered.
    expected_result = {
        "old": SaltOrderedDict([("baz", 3)]),
        "new": SaltOrderedDict([("baz", 1)]),
    }
    assert expected_result == salt.utils.data.recursive_diff(
        odict_one, odict_two, ignore_keys=["foo"]
    )


def test_dict_vs_ordereddict_ignore():
    """
    Test case comparing a dict with an SaltOrderedDict with ignore-list supplied.
    """
    dict_one = {"foo": 1, "bar": 2, "baz": 3}
    odict_two = SaltOrderedDict([("foo", 3), ("bar", 2), ("baz", 1)])
    expected_result = {"old": {"baz": 3}, "new": SaltOrderedDict([("baz", 1)])}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, odict_two, ignore_keys=["foo"]
    )


def test_mixed_nested_ignore():
    """
    Test case comparing mixed, nested items with ignore-list supplied.
    """
    dict_one = {"foo": [1], "bar": {"foo": 1, "bar": 2}, "baz": 3}
    dict_two = {"foo": [2], "bar": {"foo": 3, "bar": 2}, "baz": 1}
    expected_result = {"old": {"baz": 3}, "new": {"baz": 1}}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_keys=["foo"]
    )


def test_ordered_dict_unequal_length():
    """
    Test case comparing two SaltOrderedDicts of unequal length.
    """
    odict_one = SaltOrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
    odict_two = SaltOrderedDict([("foo", 1), ("bar", 2)])
    expected_result = {"old": SaltOrderedDict([("baz", 3)]), "new": {}}
    assert expected_result == salt.utils.data.recursive_diff(odict_one, odict_two)


def test_list_unequal_length():
    """
    Test case comparing two lists of unequal length.
    """
    list_one = [1, 2, 3]
    list_two = [1, 2, 3, 4]
    expected_result = {"old": [], "new": [4]}
    assert expected_result == salt.utils.data.recursive_diff(list_one, list_two)


def test_set_unequal_length():
    """
    Test case comparing two sets of unequal length.
    This does not do anything special, as it is unordered.
    """
    set_one = {1, 2, 3}
    set_two = {4, 3, 2, 1}
    expected_result = {"old": set(), "new": {4}}
    assert expected_result == salt.utils.data.recursive_diff(set_one, set_two)


def test_tuple_unequal_length():
    """
    Test case comparing two tuples of unequal length.
    This should be the same as comparing two ordered lists.
    """
    tuple_one = (1, 2, 3)
    tuple_two = (1, 2, 3, 4)
    expected_result = {"old": (), "new": (4,)}
    assert expected_result == salt.utils.data.recursive_diff(tuple_one, tuple_two)


def test_list_unordered():
    """
    Test case comparing two lists unordered.
    """
    list_one = [1, 2, 3, 4]
    list_two = [4, 3, 2]
    expected_result = {"old": [1], "new": []}
    assert expected_result == salt.utils.data.recursive_diff(
        list_one, list_two, ignore_order=True
    )


def test_mixed_nested_unordered():
    """
    Test case comparing nested dicts/lists unordered.
    """
    dict_one = {"foo": {"bar": [1, 2, 3]}, "bar": [{"foo": 4}, 0]}
    dict_two = {"foo": {"bar": [3, 2, 1]}, "bar": [0, {"foo": 4}]}
    expected_result = {}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_order=True
    )
    expected_result = {
        "old": {"foo": {"bar": [1, 3]}, "bar": [{"foo": 4}, 0]},
        "new": {"foo": {"bar": [3, 1]}, "bar": [0, {"foo": 4}]},
    }
    assert expected_result == salt.utils.data.recursive_diff(dict_one, dict_two)


def test_ordered_dict_unordered():
    """
    Test case comparing SaltOrderedDicts unordered.
    """
    odict_one = SaltOrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
    odict_two = SaltOrderedDict([("baz", 3), ("bar", 2), ("foo", 1)])
    expected_result = {}
    assert expected_result == salt.utils.data.recursive_diff(
        odict_one, odict_two, ignore_order=True
    )


def test_ignore_missing_keys_dict():
    """
    Test case ignoring missing keys on a comparison of dicts.
    """
    dict_one = {"foo": 1, "bar": 2, "baz": 3}
    dict_two = {"bar": 3}
    expected_result = {"old": {"bar": 2}, "new": {"bar": 3}}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_missing_keys=True
    )


def test_ignore_missing_keys_ordered_dict():
    """
    Test case not ignoring missing keys on a comparison of SaltOrderedDicts.
    """
    odict_one = SaltOrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
    odict_two = SaltOrderedDict([("bar", 3)])
    expected_result = {"old": odict_one, "new": odict_two}
    assert expected_result == salt.utils.data.recursive_diff(
        odict_one, odict_two, ignore_missing_keys=True
    )


def test_ignore_missing_keys_recursive():
    """
    Test case ignoring missing keys on a comparison of nested dicts.
    """
    dict_one = {"foo": {"bar": 2, "baz": 3}}
    dict_two = {"foo": {"baz": 3}}
    expected_result = {}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_missing_keys=True
    )
    # Compare from dict-in-dict
    dict_two = {}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_missing_keys=True
    )
    # Compare from dict-in-list
    dict_one = {"foo": ["bar", {"baz": 3}]}
    dict_two = {"foo": ["bar", {}]}
    assert expected_result == salt.utils.data.recursive_diff(
        dict_one, dict_two, ignore_missing_keys=True
    )



import os

import pytest

from salt.pillar.git_pillar import GLOBAL_ONLY, PER_REMOTE_ONLY, PER_REMOTE_OVERRIDES
from salt.utils.gitfs import GitPillar, GitPython, Pygit2
from salt.utils.immutabletypes import ImmutableDict, ImmutableList
from salt.utils.platform import get_machine_identifier as _get_machine_identifier

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.slow_test,
]


try:
    import git  # pylint: disable=unused-import

    HAS_GITPYTHON = True
except ImportError:
    HAS_GITPYTHON = False


try:
    import pygit2  # pylint: disable=unused-import

    HAS_PYGIT2 = True
except ImportError:
    HAS_PYGIT2 = False


skipif_no_gitpython = pytest.mark.skipif(not HAS_GITPYTHON, reason="Missing gitpython")
skipif_no_pygit2 = pytest.mark.skipif(not HAS_PYGIT2, reason="Missing pygit2")

testgitfs = "https://github.com/saltstack/salt-test-pillar-gitfs.git"


@pytest.fixture
def pillar_opts(salt_factories, tmp_path):
    config_defaults = {"cachedir": str(tmp_path)}
    factory = salt_factories.salt_master_daemon(
        "pillar-functional-master", defaults=config_defaults
    )
    config_defaults = dict(factory.config)
    for key, item in config_defaults.items():
        if isinstance(item, ImmutableDict):
            config_defaults[key] = dict(item)
        elif isinstance(item, ImmutableList):
            config_defaults[key] = list(item)
    return config_defaults


@pytest.fixture
def gitpython_pillar_opts(pillar_opts):
    pillar_opts["verified_git_pillar_provider"] = "gitpython"
    return pillar_opts


@pytest.fixture
def pygit2_pillar_opts(pillar_opts):
    pillar_opts["verified_git_pillar_provider"] = "pygit2"
    return pillar_opts


def _get_pillar(opts, *remotes):
    return GitPillar(
        opts,
        remotes,
        per_remote_overrides=PER_REMOTE_OVERRIDES,
        per_remote_only=PER_REMOTE_ONLY,
        global_only=GLOBAL_ONLY,
    )


@skipif_no_gitpython
def test_gitpython_pillar_provider(gitpython_pillar_opts):
    p = _get_pillar(gitpython_pillar_opts, testgitfs)
    assert len(p.remotes) == 1
    assert p.provider == "gitpython"
    assert isinstance(p.remotes[0], GitPython)


@skipif_no_pygit2
def test_pygit2_pillar_provider(pygit2_pillar_opts):
    p = _get_pillar(pygit2_pillar_opts, testgitfs)
    assert len(p.remotes) == 1
    assert p.provider == "pygit2"
    assert isinstance(p.remotes[0], Pygit2)


def _test_env(opts):
    p = _get_pillar(opts, f"__env__ {testgitfs}")
    assert len(p.remotes) == 1
    p.checkout()
    repo = p.remotes[0]
    # test that two different pillarenvs can exist at the same time
    files = set(os.listdir(repo.get_cachedir()))
    for f in (".gitignore", "README.md", "file.sls", "top.sls"):
        assert f in files
    opts["pillarenv"] = "main"
    p2 = _get_pillar(opts, f"__env__ {testgitfs}")
    assert len(p.remotes) == 1
    p2.checkout()
    repo2 = p2.remotes[0]
    files = set(os.listdir(repo2.get_cachedir()))
    for f in (".gitignore", "README.md"):
        assert f in files
    for f in ("file.sls", "top.sls", "back.sls", "rooms.sls"):
        assert f not in files
    assert repo.get_cachedir() != repo2.get_cachedir()
    files = set(os.listdir(repo.get_cachedir()))
    for f in (".gitignore", "README.md", "file.sls", "top.sls"):
        assert f in files

    # double check cache paths
    assert (
        repo.get_cache_hash() == repo2.get_cache_hash()
    )  # __env__ repos share same hash
    assert repo.get_cache_basename() != repo2.get_cache_basename()
    assert repo.get_linkdir() != repo2.get_linkdir()
    assert repo.get_salt_working_dir() != repo2.get_salt_working_dir()
    assert repo.get_cache_basename() == "master"
    assert repo2.get_cache_basename() == "main"

    assert repo.get_cache_basename() in repo.get_cachedir()
    assert (
        os.path.join(repo.get_cache_basehash(), repo.get_cache_basename())
        == repo.get_cache_full_basename()
    )
    assert repo.get_linkdir() not in repo.get_cachedir()
    assert repo.get_salt_working_dir() not in repo.get_cachedir()


@skipif_no_gitpython
def test_gitpython_env(gitpython_pillar_opts):
    _test_env(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_env(pygit2_pillar_opts):
    _test_env(pygit2_pillar_opts)


def _test_checkout_fetch_on_fail(opts):
    p = _get_pillar(opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git")
    p.checkout(fetch_on_fail=False)  # TODO write me


@skipif_no_gitpython
def test_gitpython_checkout_fetch_on_fail(gitpython_pillar_opts):
    _test_checkout_fetch_on_fail(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_checkout_fetch_on_fail(pygit2_pillar_opts):
    _test_checkout_fetch_on_fail(pygit2_pillar_opts)


def _test_multiple_repos(opts):
    p = _get_pillar(
        opts,
        f"__env__ {testgitfs}",
        f"main {testgitfs}",
        f"branch {testgitfs}",
        "__env__ https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
        "other https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
    )
    p.checkout()
    assert len(p.remotes) == 5
    # make sure all repos dont share cache and working dir
    assert len({r.get_cachedir() for r in p.remotes}) == 5
    assert len({r.get_salt_working_dir() for r in p.remotes}) == 5

    p2 = _get_pillar(
        opts,
        f"__env__ {testgitfs}",
        f"main {testgitfs}",
        f"branch {testgitfs}",
        "__env__ https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
        "other https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
    )
    p2.checkout()
    assert len(p2.remotes) == 5
    # make sure that repos are given same cache dir
    for repo, repo2 in zip(p.remotes, p2.remotes):
        assert repo.get_cachedir() == repo2.get_cachedir()
        assert repo.get_salt_working_dir() == repo2.get_salt_working_dir()
    opts["pillarenv"] = "main"
    p3 = _get_pillar(
        opts,
        f"__env__ {testgitfs}",
        f"main {testgitfs}",
        f"branch {testgitfs}",
        "__env__ https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
        "other https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
    )
    p3.checkout()
    # check that __env__ has different cache with different pillarenv
    assert p.remotes[0].get_cachedir() != p3.remotes[0].get_cachedir()
    assert p.remotes[1].get_cachedir() == p3.remotes[1].get_cachedir()
    assert p.remotes[2].get_cachedir() == p3.remotes[2].get_cachedir()
    assert p.remotes[3].get_cachedir() != p3.remotes[3].get_cachedir()
    assert p.remotes[4].get_cachedir() == p3.remotes[4].get_cachedir()

    # check that other branch data is in cache
    files = set(os.listdir(p.remotes[4].get_cachedir()))
    for f in (".gitignore", "README.md", "file.sls", "top.sls", "other_env.sls"):
        assert f in files


@skipif_no_gitpython
def test_gitpython_multiple_repos(gitpython_pillar_opts):
    _test_multiple_repos(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_multiple_repos(pygit2_pillar_opts):
    _test_multiple_repos(pygit2_pillar_opts)


def _test_fetch_request(opts):
    p = _get_pillar(
        opts,
        f"__env__ {testgitfs}",
        "other https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
    )
    frequest = os.path.join(p.remotes[0].get_salt_working_dir(), "fetch_request")
    frequest_other = os.path.join(p.remotes[1].get_salt_working_dir(), "fetch_request")
    opts["pillarenv"] = "main"
    p2 = _get_pillar(opts, f"__env__ {testgitfs}")
    frequest2 = os.path.join(p2.remotes[0].get_salt_working_dir(), "fetch_request")
    assert frequest != frequest2
    assert os.path.isfile(frequest) is False
    assert os.path.isfile(frequest2) is False
    assert os.path.isfile(frequest_other) is False
    p.fetch_remotes()
    assert os.path.isfile(frequest) is False
    # fetch request was placed
    assert os.path.isfile(frequest2) is True
    p2.checkout()
    # fetch request was found
    assert os.path.isfile(frequest2) is False
    p2.fetch_remotes()
    assert os.path.isfile(frequest) is True
    assert os.path.isfile(frequest2) is False
    assert os.path.isfile(frequest_other) is False
    for _ in range(3):
        p2.fetch_remotes()
    assert os.path.isfile(frequest) is True
    assert os.path.isfile(frequest2) is False
    assert os.path.isfile(frequest_other) is False
    # fetch request should still be processed even on fetch_on_fail=False
    p.checkout(fetch_on_fail=False)
    assert os.path.isfile(frequest) is False
    assert os.path.isfile(frequest2) is False
    assert os.path.isfile(frequest_other) is False


@skipif_no_gitpython
def test_gitpython_fetch_request(gitpython_pillar_opts):
    _test_fetch_request(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_fetch_request(pygit2_pillar_opts):
    _test_fetch_request(pygit2_pillar_opts)


def _test_clear_old_remotes(opts):
    p = _get_pillar(
        opts,
        f"__env__ {testgitfs}",
        "other https://github.com/saltstack/salt-test-pillar-gitfs-2.git",
    )
    repo = p.remotes[0]
    repo2 = p.remotes[1]
    opts["pillarenv"] = "main"
    p2 = _get_pillar(opts, f"__env__ {testgitfs}")
    repo3 = p2.remotes[0]
    assert os.path.isdir(repo.get_cachedir()) is True
    assert os.path.isdir(repo2.get_cachedir()) is True
    assert os.path.isdir(repo3.get_cachedir()) is True
    p.clear_old_remotes()
    assert os.path.isdir(repo.get_cachedir()) is True
    assert os.path.isdir(repo2.get_cachedir()) is True
    assert os.path.isdir(repo3.get_cachedir()) is True
    p2.clear_old_remotes()
    assert os.path.isdir(repo.get_cachedir()) is True
    assert os.path.isdir(repo2.get_cachedir()) is False
    assert os.path.isdir(repo3.get_cachedir()) is True


@skipif_no_gitpython
def test_gitpython_clear_old_remotes(gitpython_pillar_opts):
    _test_clear_old_remotes(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_clear_old_remotes(pygit2_pillar_opts):
    _test_clear_old_remotes(pygit2_pillar_opts)


def _test_remote_map(opts):
    p = _get_pillar(
        opts,
        testgitfs,
    )
    p.fetch_remotes()
    assert len(p.remotes) == 1
    assert os.path.isfile(
        os.path.join(opts["cachedir"], "git_pillar", "remote_map.txt")
    )


@skipif_no_gitpython
def test_gitpython_remote_map(gitpython_pillar_opts):
    _test_remote_map(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_remote_map(pygit2_pillar_opts):
    _test_remote_map(pygit2_pillar_opts)


def _test_lock(opts):
    p = _get_pillar(
        opts,
        testgitfs,
    )
    p.fetch_remotes()
    assert len(p.remotes) == 1
    repo = p.remotes[0]
    mach_id = _get_machine_identifier().get("machine_id", "no_machine_id_available")
    assert repo.get_salt_working_dir() in repo._get_lock_file()
    assert repo.lock() == (
        [
            (
                f"Set update lock for git_pillar remote '{testgitfs}' on machine_id '{mach_id}'"
            )
        ],
        [],
    )
    assert os.path.isfile(repo._get_lock_file())
    assert repo.clear_lock() == (
        [
            (
                f"Removed update lock for git_pillar remote '{testgitfs}' on machine_id '{mach_id}'"
            )
        ],
        [],
    )
    assert not os.path.isfile(repo._get_lock_file())


@skipif_no_gitpython
def test_gitpython_lock(gitpython_pillar_opts):
    _test_lock(gitpython_pillar_opts)


@skipif_no_pygit2
def test_pygit2_lock(pygit2_pillar_opts):
    _test_lock(pygit2_pillar_opts)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.apache as apache
import salt.utils.files
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {apache: {}}


def test_configfile():
    """
    Test to allows for inputting a yaml dictionary into a file
    for apache configuration files.
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        name = "/etc/distro/specific/apache.conf"
        config = 'VirtualHost: this: "*:80"'
        new_config = 'LiteralHost: that: "*:79"'

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        with patch.object(salt.utils.files, "fopen", mock_open(read_data=config)):
            mock_config = MagicMock(return_value=config)
            with patch.dict(apache.__salt__, {"apache.config": mock_config}):
                ret.update({"comment": "Configuration is up to date."})
                assert apache.configfile(name, config) == ret

        with patch.object(salt.utils.files, "fopen", mock_open(read_data=config)):
            mock_config = MagicMock(return_value=new_config)
            with patch.dict(apache.__salt__, {"apache.config": mock_config}):
                ret.update(
                    {
                        "comment": "Configuration will update.",
                        "changes": {"new": new_config, "old": config},
                        "result": None,
                    }
                )
                with patch.dict(apache.__opts__, {"test": True}):
                    assert apache.configfile(name, new_config) == ret

        with patch.object(salt.utils.files, "fopen", mock_open(read_data=config)):
            mock_config = MagicMock(return_value=new_config)
            with patch.dict(apache.__salt__, {"apache.config": mock_config}):
                ret.update(
                    {"comment": "Successfully created configuration.", "result": True}
                )
                with patch.dict(apache.__opts__, {"test": False}):
                    assert apache.configfile(name, config) == ret



"""
:codeauthor: Herbert Buurman <herbert.buurman@ogd.nl>
"""

import pytest

import salt.states.acme as acme
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {acme: {"__opts__": {"test": False}}}


def test_cert_no_changes_test():
    """
    Test cert state with no needed changes. (test=True)
    """
    with patch.dict(
        acme.__salt__,
        {
            "acme.has": MagicMock(return_value=True),
            "acme.needs_renewal": MagicMock(return_value=False),
        },
    ), patch.dict(acme.__opts__, {"test": True}):
        match = {
            "name": "test",
            "result": True,
            "comment": ["Certificate test exists and does not need renewal."],
            "changes": {},
        }
        assert acme.cert("test") == match
        assert acme.cert("testing.example.com", certname="test") == match


def test_cert_no_changes():
    """
    Test cert state with no needed changes. (test=False)
    """
    with patch.dict(
        acme.__salt__,
        {
            "acme.has": MagicMock(return_value=True),
            "acme.needs_renewal": MagicMock(return_value=False),
        },
    ):
        match = {
            "name": "test",
            "result": True,
            "comment": ["Certificate test exists and does not need renewal."],
            "changes": {},
        }
        assert acme.cert("test") == match
        assert acme.cert("testing.example.com", certname="test") == match


def test_cert_fresh_certificate_test():
    """
    Test cert state fetching a new certificate. (test=True)
    """
    # With test=True
    with patch.dict(
        acme.__salt__,
        {
            "acme.has": MagicMock(return_value=False),
            "acme.info": MagicMock(return_value={"foo": "bar"}),
        },
    ), patch.dict(acme.__opts__, {"test": True}):
        match = {
            "name": "test",
            "result": None,
            "comment": ["Certificate test would have been obtained."],
            "changes": {"old": "current certificate", "new": "new certificate"},
        }
        assert acme.cert("test") == match
        assert acme.cert("testing.example.com", certname="test") == match


def test_cert_fresh_certificate():
    """
    Test cert state fetching a new certificate. (test=False)
    """
    with patch.dict(
        acme.__salt__,
        {
            "acme.has": MagicMock(return_value=False),
            "acme.cert": MagicMock(return_value={"result": True, "comment": "Mockery"}),
            "acme.info": MagicMock(return_value={"foo": "bar"}),
        },
    ):
        match = {
            "name": "test",
            "result": True,
            "comment": ["Mockery"],
            "changes": {"new": {"foo": "bar"}},
        }
        assert acme.cert("test") == match
        assert acme.cert("testing.example.com", certname="test") == match


def test_cert_renew_certificate_test():
    """
    Test cert state renewing a certificate. (test=True)
    """
    with patch.dict(
        acme.__salt__,
        {
            "acme.has": MagicMock(return_value=True),
            "acme.needs_renewal": MagicMock(return_value=True),
            "acme.info": MagicMock(side_effect=[{"name": "old cert"}] * 2),
            "acme.cert": MagicMock(return_value={"result": True, "comment": "Mockery"}),
        },
    ), patch.dict(acme.__opts__, {"test": True}):
        match = {
            "name": "test",
            "result": None,
            "comment": ["Certificate test would have been renewed."],
            "changes": {"old": "current certificate", "new": "new certificate"},
        }
        assert acme.cert("test") == match
        assert acme.cert("testing.example.com", certname="test") == match


def test_cert_renew_certificate():
    """
    Test cert state renewing a certificate. (test=False)
    """
    with patch.dict(
        acme.__salt__,
        {  # pylint: disable=no-member
            "acme.has": MagicMock(return_value=True),
            "acme.needs_renewal": MagicMock(return_value=True),
            "acme.info": MagicMock(
                side_effect=[{"name": "old cert"}, {"name": "new cert"}] * 2
            ),
            "acme.cert": MagicMock(return_value={"result": True, "comment": "Mockery"}),
        },
    ):
        match = {
            "name": "test",
            "result": True,
            "comment": ["Mockery"],
            "changes": {"old": {"name": "old cert"}, "new": {"name": "new cert"}},
        }
        assert acme.cert("test") == match
        assert acme.cert("testing.example.com", certname="test") == match



import pytest

import salt.modules.cmdmod as cmdmod
from salt.exceptions import CommandExecutionError

pytestmark = [pytest.mark.windows_whitelisted]


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.fixture(scope="module")
def configure_loader_modules():
    return {
        cmdmod: {
            "__grains__": {"os": "linux", "os_family": "linux"},
        }
    }


@pytest.mark.skip_on_windows
@pytest.mark.skip_if_not_root
def test_runas(account):
    ret = cmdmod.run("id", runas=account.username)
    assert f"gid={account.info.gid}" in ret
    assert f"uid={account.info.uid}" in ret


def test_runas_missing_user():
    with pytest.raises(CommandExecutionError):
        cmdmod.run("echo HOLO", runas="non-existent-user", password="junk")



# Majority of code shamelessly stolen from
# http://www.v13.gr/blog/?p=303
"""
Authenticate via a PKI certificate.

.. note::

    This module is Experimental and should be used with caution

Provides an authenticate function that will allow the caller to authenticate
a user via their public cert against a pre-defined Certificate Authority.

TODO: Add a 'ca_dir' option to configure a directory of CA files, a la Apache.

:depends:    - pyOpenSSL module
"""
import logging

import salt.utils.files
import salt.utils.versions

# pylint: disable=import-error
try:
    try:
        from M2Crypto import X509

        HAS_M2 = True
    except ImportError:
        HAS_M2 = False
        try:
            from Cryptodome.Util import asn1
        except ImportError:
            from Crypto.Util import asn1  # nosec
        import OpenSSL  # pylint: disable=W8410
    HAS_DEPS = True
except ImportError:
    HAS_DEPS = False
# pylint: enable=import-error


log = logging.getLogger(__name__)


def __virtual__():
    """
    Requires newer pycrypto and pyOpenSSL
    """
    if HAS_DEPS:
        return True
    return False


def auth(username, password, **kwargs):
    """
    Returns True if the given user cert (password is the cert contents)
    was issued by the CA and if cert's Common Name is equal to username.

    Returns False otherwise.

    ``username``: we need it to run the auth function from CLI/API;
                  it should be in master config auth/acl
    ``password``: contents of user certificate (pem-encoded user public key);
                  why "password"? For CLI, it's the only available name

    Configure the CA cert in the master config file:

    .. code-block:: yaml

        external_auth:
          pki:
            ca_file: /etc/pki/tls/ca_certs/trusted-ca.crt
            your_user:
              - .*
    """
    salt.utils.versions.warn_until(
        "Argon",
        "This module has been deprecated as it is known to be insecure.",
    )
    pem = password
    cacert_file = __salt__["config.get"]("external_auth:pki:ca_file")

    log.debug("Attempting to authenticate via pki.")
    log.debug("Using CA file: %s", cacert_file)
    log.debug("Certificate contents: %s", pem)

    if HAS_M2:
        cert = X509.load_cert_string(pem, X509.FORMAT_PEM)
        cacert = X509.load_cert(cacert_file, X509.FORMAT_PEM)
        if cert.verify(cacert.get_pubkey()):
            log.info("Successfully authenticated certificate: %s", pem)
            return True
        log.info("Failed to authenticate certificate: %s", pem)
        return False

    c = OpenSSL.crypto  # pylint: disable=used-before-assignment
    cert = c.load_certificate(c.FILETYPE_PEM, pem)

    with salt.utils.files.fopen(cacert_file) as f:
        cacert = c.load_certificate(c.FILETYPE_PEM, f.read())

    # Get the signing algorithm
    algo = cert.get_signature_algorithm()

    # Get the ASN1 format of the certificate
    cert_asn1 = c.dump_certificate(c.FILETYPE_ASN1, cert)

    # Decode the certificate
    der = asn1.DerSequence()  # pylint: disable=used-before-assignment
    der.decode(cert_asn1)

    # The certificate has three parts:
    # - certificate
    # - signature algorithm
    # - signature
    # http://usefulfor.com/nothing/2009/06/10/x509-certificate-basics/
    der_cert = der[0]
    # der_algo = der[1]
    der_sig = der[2]

    # The signature is a BIT STRING (Type 3)
    # Decode that as well
    der_sig_in = asn1.DerObject()
    der_sig_in.decode(der_sig)

    # Get the payload
    sig0 = der_sig_in.payload

    # Do the following to see a validation error for tests
    # der_cert=der_cert[:20]+'1'+der_cert[21:]

    # First byte is the number of unused bits. This should be 0
    # http://msdn.microsoft.com/en-us/library/windows/desktop/bb540792(v=vs.85).aspx
    if sig0[0] != "\x00":
        raise Exception("Number of unused bits is strange")
    # Now get the signature itself
    sig = sig0[1:]

    # And verify the certificate
    try:
        c.verify(cacert, sig, der_cert, algo)
        assert (
            dict(cert.get_subject().get_components())["CN"] == username
        ), "Certificate's CN should match the username"
        log.info("Successfully authenticated certificate: %s", pem)
        return True
    except (OpenSSL.crypto.Error, AssertionError):
        log.info("Failed to authenticate certificate: %s", pem)
    return False
